# -*- coding: utf-8; -*-
#
# Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
# license agreements.  See the NOTICE file distributed with this work for
# additional information regarding copyright ownership.  Crate licenses
# this file to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may
# obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
# However, if you have executed another commercial license agreement
# with Crate these terms will supersede the license and you may use the
# software solely pursuant to the terms of the relevant commercial agreement.

from datetime import datetime
from unittest import TestCase
from unittest.mock import patch, MagicMock

from crate.client.sqlalchemy.types import ObjectType

import sqlalchemy as sa
from sqlalchemy.orm import Session
try:
    from sqlalchemy.orm import declarative_base
except ImportError:
    from sqlalchemy.ext.declarative import declarative_base

from crate.client.cursor import Cursor


fake_cursor = MagicMock(name='fake_cursor')
fake_cursor.rowcount = 1
FakeCursor = MagicMock(name='FakeCursor', spec=Cursor)
FakeCursor.return_value = fake_cursor


class SqlAlchemyUpdateTest(TestCase):

    def setUp(self):
        self.engine = sa.create_engine('crate://')
        self.base = declarative_base()

        class Character(self.base):
            __tablename__ = 'characters'

            name = sa.Column(sa.String, primary_key=True)
            age = sa.Column(sa.Integer)
            obj = sa.Column(ObjectType)
            ts = sa.Column(sa.DateTime, onupdate=datetime.utcnow)

        self.character = Character
        self.session = Session(bind=self.engine)

    @patch('crate.client.connection.Cursor', FakeCursor)
    def test_onupdate_is_triggered(self):
        char = self.character(name='Arthur')
        self.session.add(char)
        self.session.commit()
        now = datetime.utcnow()

        fake_cursor.fetchall.return_value = [('Arthur', None)]
        fake_cursor.description = (
            ('characters_name', None, None, None, None, None, None),
            ('characters_ts', None, None, None, None, None, None),
        )

        char.age = 40
        self.session.commit()

        expected_stmt = ("UPDATE characters SET age = ?, "
                         "ts = ? WHERE characters.name = ?")
        args, kwargs = fake_cursor.execute.call_args
        stmt = args[0]
        args = args[1]
        self.assertEqual(expected_stmt, stmt)
        self.assertEqual(40, args[0])
        dt = datetime.strptime(args[1], '%Y-%m-%dT%H:%M:%S.%fZ')
        self.assertIsInstance(dt, datetime)
        self.assertGreater(dt, now)
        self.assertEqual('Arthur', args[2])

    @patch('crate.client.connection.Cursor', FakeCursor)
    def test_bulk_update(self):
        """
            Checks whether bulk updates work correctly
            on native types and Crate types.
        """
        before_update_time = datetime.utcnow()

        self.session.query(self.character).update({
            # change everyone's name to Julia
            self.character.name: 'Julia',
            self.character.obj: {'favorite_book': 'Romeo & Juliet'}
        })

        self.session.commit()

        expected_stmt = ("UPDATE characters SET "
                         "name = ?, obj = ?, ts = ?")
        args, kwargs = fake_cursor.execute.call_args
        stmt = args[0]
        args = args[1]
        self.assertEqual(expected_stmt, stmt)
        self.assertEqual('Julia', args[0])
        self.assertEqual({'favorite_book': 'Romeo & Juliet'}, args[1])
        dt = datetime.strptime(args[2], '%Y-%m-%dT%H:%M:%S.%fZ')
        self.assertIsInstance(dt, datetime)
        self.assertGreater(dt, before_update_time)



# -*- coding: utf-8; -*-
#
# Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
# license agreements.  See the NOTICE file distributed with this work for
# additional information regarding copyright ownership.  Crate licenses
# this file to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may
# obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
# However, if you have executed another commercial license agreement
# with Crate these terms will supersede the license and you may use the
# software solely pursuant to the terms of the relevant commercial agreement.
import math
import sys
from unittest import TestCase, skipIf
from unittest.mock import patch, MagicMock

import sqlalchemy as sa
from sqlalchemy.orm import Session

from crate.client.sqlalchemy.sa_version import SA_VERSION, SA_2_0, SA_1_4

try:
    from sqlalchemy.orm import declarative_base
except ImportError:
    from sqlalchemy.ext.declarative import declarative_base

from crate.client.cursor import Cursor


fake_cursor = MagicMock(name='fake_cursor')
FakeCursor = MagicMock(name='FakeCursor', spec=Cursor, return_value=fake_cursor)


class SqlAlchemyBulkTest(TestCase):

    def setUp(self):
        self.engine = sa.create_engine('crate://')
        Base = declarative_base()

        class Character(Base):
            __tablename__ = 'characters'

            name = sa.Column(sa.String, primary_key=True)
            age = sa.Column(sa.Integer)

        self.character = Character
        self.session = Session(bind=self.engine)

    @skipIf(SA_VERSION >= SA_2_0, "SQLAlchemy 2.x uses modern bulk INSERT mode")
    @patch('crate.client.connection.Cursor', FakeCursor)
    def test_bulk_save_legacy(self):
        """
        Verify legacy SQLAlchemy bulk INSERT mode.

        > bulk_save_objects: Perform a bulk save of the given list of objects.
        > This method is a legacy feature as of the 2.0 series of SQLAlchemy. For modern
        > bulk INSERT and UPDATE, see the sections ORM Bulk INSERT Statements and ORM Bulk
        > UPDATE by Primary Key.
        >
        > -- https://docs.sqlalchemy.org/orm/session_api.html#sqlalchemy.orm.Session.bulk_save_objects

        > The Session includes legacy methods for performing "bulk" INSERT and UPDATE
        > statements. These methods share implementations with the SQLAlchemy 2.0
        > versions of these features, described at ORM Bulk INSERT Statements and
        > ORM Bulk UPDATE by Primary Key, however lack many features, namely RETURNING
        > support as well as support for session-synchronization.
        >
        > -- https://docs.sqlalchemy.org/orm/queryguide/dml.html#legacy-session-bulk-insert-methods

        > The 1.4 version of the "ORM bulk insert" methods are really not very efficient and
        > don't grant that much of a performance bump vs. regular ORM `session.add()`, provided
        > in both cases the objects you provide already have their primary key values assigned.
        > SQLAlchemy 2.0 made a much more comprehensive change to how this all works as well so
        > that all INSERT methods are essentially extremely fast now, relative to the 1.x series.
        >
        > -- https://github.com/sqlalchemy/sqlalchemy/discussions/6935#discussioncomment-4789701
        """
        chars = [
            self.character(name='Arthur', age=35),
            self.character(name='Banshee', age=26),
            self.character(name='Callisto', age=37),
        ]

        fake_cursor.description = ()
        fake_cursor.rowcount = len(chars)
        fake_cursor.executemany.return_value = [
            {'rowcount': 1},
            {'rowcount': 1},
            {'rowcount': 1},
        ]
        self.session.bulk_save_objects(chars)
        (stmt, bulk_args), _ = fake_cursor.executemany.call_args

        expected_stmt = "INSERT INTO characters (name, age) VALUES (?, ?)"
        self.assertEqual(expected_stmt, stmt)

        expected_bulk_args = (
            ('Arthur', 35),
            ('Banshee', 26),
            ('Callisto', 37)
        )
        self.assertSequenceEqual(expected_bulk_args, bulk_args)

    @skipIf(SA_VERSION < SA_2_0, "SQLAlchemy 1.x uses legacy bulk INSERT mode")
    @patch('crate.client.connection.Cursor', FakeCursor)
    def test_bulk_save_modern(self):
        """
        Verify modern SQLAlchemy bulk INSERT mode.

        > A list of parameter dictionaries sent to the `Session.execute.params` parameter,
        > separate from the Insert object itself, will invoke *bulk INSERT mode* for the
        > statement, which essentially means the operation will optimize as much as
        > possible for many rows.
        >
        > -- https://docs.sqlalchemy.org/orm/queryguide/dml.html#orm-queryguide-bulk-insert

        > We have been looking into getting performance optimizations
        > from `bulk_save()` to be inherently part of `add_all()`.
        >
        > -- https://github.com/sqlalchemy/sqlalchemy/discussions/6935#discussioncomment-1233465

        > The remaining performance limitation, that the `cursor.executemany()` DBAPI method
        > does not allow for rows to be fetched, is resolved for most backends by *foregoing*
        > the use of `executemany()` and instead restructuring individual INSERT statements
        > to each accommodate a large number of rows in a single statement that is invoked
        > using `cursor.execute()`. This approach originates from the `psycopg2` fast execution
        > helpers feature of the `psycopg2` DBAPI, which SQLAlchemy incrementally added more
        > and more support towards in recent release series.
        >
        > -- https://docs.sqlalchemy.org/core/connections.html#engine-insertmanyvalues
        """

        # Don't truncate unittest's diff output on `assertListEqual`.
        self.maxDiff = None

        chars = [
            self.character(name='Arthur', age=35),
            self.character(name='Banshee', age=26),
            self.character(name='Callisto', age=37),
        ]

        fake_cursor.description = ()
        fake_cursor.rowcount = len(chars)
        fake_cursor.execute.return_value = [
            {'rowcount': 1},
            {'rowcount': 1},
            {'rowcount': 1},
        ]
        self.session.add_all(chars)
        self.session.commit()
        (stmt, bulk_args), _ = fake_cursor.execute.call_args

        expected_stmt = "INSERT INTO characters (name, age) VALUES (?, ?), (?, ?), (?, ?)"
        self.assertEqual(expected_stmt, stmt)

        expected_bulk_args = (
            'Arthur', 35,
            'Banshee', 26,
            'Callisto', 37,
        )
        self.assertSequenceEqual(expected_bulk_args, bulk_args)

    @skipIf(sys.version_info < (3, 8), "SQLAlchemy/pandas is not supported on Python <3.8")
    @skipIf(SA_VERSION < SA_1_4, "SQLAlchemy 1.3 is not supported by pandas")
    @patch('crate.client.connection.Cursor', mock_cursor=FakeCursor)
    def test_bulk_save_pandas(self, mock_cursor):
        """
        Verify bulk INSERT with pandas.
        """
        from pandas._testing import makeTimeDataFrame
        from crate.client.sqlalchemy.support import insert_bulk

        # 42 records / 8 chunksize = 5.25, which means 6 batches will be emitted.
        INSERT_RECORDS = 42
        CHUNK_SIZE = 8
        OPCOUNT = math.ceil(INSERT_RECORDS / CHUNK_SIZE)

        # Create a DataFrame to feed into the database.
        df = makeTimeDataFrame(nper=INSERT_RECORDS, freq="S")

        dburi = "crate://localhost:4200"
        engine = sa.create_engine(dburi, echo=True)
        retval = df.to_sql(
            name="test-testdrive",
            con=engine,
            if_exists="replace",
            index=False,
            chunksize=CHUNK_SIZE,
            method=insert_bulk,
        )
        self.assertIsNone(retval)

        # Initializing the query has an overhead of two calls to the cursor object, probably one
        # initial connection from the DB-API driver, to inquire the database version, and another
        # one, for SQLAlchemy. SQLAlchemy will use it to inquire the table schema using `information_schema`,
        # and to eventually issue the `CREATE TABLE ...` statement.
        effective_op_count = mock_cursor.call_count - 2

        # Verify number of batches.
        self.assertEqual(effective_op_count, OPCOUNT)

    @skipIf(sys.version_info < (3, 8), "SQLAlchemy/Dask is not supported on Python <3.8")
    @skipIf(SA_VERSION < SA_1_4, "SQLAlchemy 1.3 is not supported by pandas")
    @patch('crate.client.connection.Cursor', mock_cursor=FakeCursor)
    def test_bulk_save_dask(self, mock_cursor):
        """
        Verify bulk INSERT with Dask.
        """
        import dask.dataframe as dd
        from pandas._testing import makeTimeDataFrame
        from crate.client.sqlalchemy.support import insert_bulk

        # 42 records / 4 partitions means each partition has a size of 10.5 elements.
        # Because the chunk size 8 is slightly smaller than 10, the partition will not
        # fit into it, so two batches will be emitted to the database for each data
        # partition. 4 partitions * 2 batches = 8 insert operations will be emitted.
        # Those settings are a perfect example of non-optimal settings, and have been
        # made so on purpose, in order to demonstrate that using optimal settings
        # is crucial.
        INSERT_RECORDS = 42
        NPARTITIONS = 4
        CHUNK_SIZE = 8
        OPCOUNT = math.ceil(INSERT_RECORDS / NPARTITIONS / CHUNK_SIZE) * NPARTITIONS

        # Create a DataFrame to feed into the database.
        df = makeTimeDataFrame(nper=INSERT_RECORDS, freq="S")
        ddf = dd.from_pandas(df, npartitions=NPARTITIONS)

        dburi = "crate://localhost:4200"
        retval = ddf.to_sql(
            name="test-testdrive",
            uri=dburi,
            if_exists="replace",
            index=False,
            chunksize=CHUNK_SIZE,
            method=insert_bulk,
            parallel=True,
        )
        self.assertIsNone(retval)

        # Each of the insert operation incurs another call to the cursor object. This is probably
        # the initial connection from the DB-API driver, to inquire the database version.
        # This compensation formula has been determined empirically / by educated guessing.
        effective_op_count = (mock_cursor.call_count - 2 * NPARTITIONS) - 2

        # Verify number of batches.
        self.assertEqual(effective_op_count, OPCOUNT)



# -*- coding: utf-8; -*-
#
# Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
# license agreements.  See the NOTICE file distributed with this work for
# additional information regarding copyright ownership.  Crate licenses
# this file to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may
# obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
# However, if you have executed another commercial license agreement
# with Crate these terms will supersede the license and you may use the
# software solely pursuant to the terms of the relevant commercial agreement.
import unittest


class ClientMocked(object):

    active_servers = ["http://localhost:4200"]

    def __init__(self):
        self.response = {}
        self._server_infos = ("http://localhost:4200", "my server", "2.0.0")

    def sql(self, stmt=None, parameters=None, bulk_parameters=None):
        return self.response

    def server_infos(self, server):
        return self._server_infos

    def set_next_response(self, response):
        self.response = response

    def set_next_server_infos(self, server, server_name, version):
        self._server_infos = (server, server_name, version)

    def close(self):
        pass


class ParametrizedTestCase(unittest.TestCase):
    """
    TestCase classes that want to be parametrized should
    inherit from this class.

    https://eli.thegreenplace.net/2011/08/02/python-unit-testing-parametrized-test-cases
    """
    def __init__(self, methodName="runTest", param=None):
        super(ParametrizedTestCase, self).__init__(methodName)
        self.param = param

    @staticmethod
    def parametrize(testcase_klass, param=None):
        """ Create a suite containing all tests taken from the given
            subclass, passing them the parameter 'param'.
        """
        testloader = unittest.TestLoader()
        testnames = testloader.getTestCaseNames(testcase_klass)
        suite = unittest.TestSuite()
        for name in testnames:
            suite.addTest(testcase_klass(name, param=param))
        return suite



# -*- coding: utf-8; -*-
#
# Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
# license agreements.  See the NOTICE file distributed with this work for
# additional information regarding copyright ownership.  Crate licenses
# this file to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may
# obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
# However, if you have executed another commercial license agreement
# with Crate these terms will supersede the license and you may use the
# software solely pursuant to the terms of the relevant commercial agreement.


class Error(Exception):

    def __init__(self, msg=None, error_trace=None):
        # for compatibility reasons we want to keep the exception message
        # attribute because clients may depend on it
        if msg:
            self.message = msg
        super(Error, self).__init__(msg)
        self.error_trace = error_trace


class Warning(Exception):
    pass


class InterfaceError(Error):
    pass


class DatabaseError(Error):
    pass


class InternalError(DatabaseError):
    pass


class OperationalError(DatabaseError):
    pass


class ProgrammingError(DatabaseError):
    pass


class IntegrityError(DatabaseError):
    pass


class DataError(DatabaseError):
    pass


class NotSupportedError(DatabaseError):
    pass


# exceptions not in db api

class ConnectionError(OperationalError):
    pass


class BlobException(Exception):
    def __init__(self, table, digest):
        self.table = table
        self.digest = digest

    def __str__(self):
        return "{table}/{digest}".format(table=self.table, digest=self.digest)


class DigestNotFoundException(BlobException):
    pass


class BlobLocationNotFoundException(BlobException):
    pass


class TimezoneUnawareException(Error):
    pass



# -*- coding: utf-8; -*-
#
# Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
# license agreements.  See the NOTICE file distributed with this work for
# additional information regarding copyright ownership.  Crate licenses
# this file to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may
# obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
# However, if you have executed another commercial license agreement
# with Crate these terms will supersede the license and you may use the
# software solely pursuant to the terms of the relevant commercial agreement.

from sqlalchemy.sql.expression import ColumnElement, literal
from sqlalchemy.ext.compiler import compiles


class Match(ColumnElement):
    inherit_cache = True

    def __init__(self, column, term, match_type=None, options=None):
        super(Match, self).__init__()
        self.column = column
        self.term = term
        self.match_type = match_type
        self.options = options

    def compile_column(self, compiler):
        if isinstance(self.column, dict):
            column = ', '.join(
                sorted(["{0} {1}".format(compiler.process(k), v)
                       for k, v in self.column.items()])
            )
            return "({0})".format(column)
        else:
            return "{0}".format(compiler.process(self.column))

    def compile_term(self, compiler):
        return compiler.process(literal(self.term))

    def compile_using(self, compiler):
        if self.match_type:
            using = "using {0}".format(self.match_type)
            with_clause = self.with_clause()
            if with_clause:
                using = ' '.join([using, with_clause])
            return using
        if self.options:
            raise ValueError("missing match_type. " +
                             "It's not allowed to specify options " +
                             "without match_type")

    def with_clause(self):
        if self.options:
            options = ', '.join(
                sorted(["{0}={1}".format(k, v)
                       for k, v in self.options.items()])
            )

            return "with ({0})".format(options)


def match(column, term, match_type=None, options=None):
    """Generates match predicate for fulltext search

    :param column: A reference to a column or an index, or a subcolumn, or a
     dictionary of subcolumns with boost values.

    :param term: The term to match against. This string is analyzed and the
     resulting tokens are compared to the index.

    :param match_type (optional): The match type. Determine how the term is
     applied and the score calculated.

    :param options (optional): The match options. Specify match type behaviour.
     (Not possible without a specified match type.) Match options must be
     supplied as a dictionary.
    """
    return Match(column, term, match_type, options)


@compiles(Match)
def compile_match(match, compiler, **kwargs):
    func = "match(%s, %s)" % (
        match.compile_column(compiler),
        match.compile_term(compiler)
    )
    using = match.compile_using(compiler)
    if using:
        func = ' '.join([func, using])
    return func



# -*- coding: utf-8; -*-
#
# Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
# license agreements.  See the NOTICE file distributed with this work for
# additional information regarding copyright ownership.  Crate licenses
# this file to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may
# obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
# However, if you have executed another commercial license agreement
# with Crate these terms will supersede the license and you may use the
# software solely pursuant to the terms of the relevant commercial agreement.


from unittest import TestCase
from unittest.mock import MagicMock

import sqlalchemy as sa
from sqlalchemy.orm import Session
try:
    from sqlalchemy.orm import declarative_base
except ImportError:
    from sqlalchemy.ext.declarative import declarative_base

from crate.client.sqlalchemy.types import ObjectType
from crate.client.sqlalchemy.predicates import match
from crate.client.cursor import Cursor


fake_cursor = MagicMock(name='fake_cursor')
FakeCursor = MagicMock(name='FakeCursor', spec=Cursor)
FakeCursor.return_value = fake_cursor


class SqlAlchemyMatchTest(TestCase):

    def setUp(self):
        self.engine = sa.create_engine('crate://')
        metadata = sa.MetaData()
        self.quotes = sa.Table('quotes', metadata,
                               sa.Column('author', sa.String),
                               sa.Column('quote', sa.String))
        self.session, self.Character = self.set_up_character_and_session()
        self.maxDiff = None

    def assertSQL(self, expected_str, actual_expr):
        self.assertEqual(expected_str, str(actual_expr).replace('\n', ''))

    def set_up_character_and_session(self):
        Base = declarative_base()

        class Character(Base):
            __tablename__ = 'characters'
            name = sa.Column(sa.String, primary_key=True)
            info = sa.Column(ObjectType)

        session = Session(bind=self.engine)
        return session, Character

    def test_simple_match(self):
        query = self.session.query(self.Character.name) \
                    .filter(match(self.Character.name, 'Trillian'))
        self.assertSQL(
            "SELECT characters.name AS characters_name FROM characters " +
            "WHERE match(characters.name, ?)",
            query
        )

    def test_match_boost(self):
        query = self.session.query(self.Character.name) \
            .filter(match({self.Character.name: 0.5}, 'Trillian'))
        self.assertSQL(
            "SELECT characters.name AS characters_name FROM characters " +
            "WHERE match((characters.name 0.5), ?)",
            query
        )

    def test_muli_match(self):
        query = self.session.query(self.Character.name) \
            .filter(match({self.Character.name: 0.5,
                           self.Character.info['race']: 0.9},
                          'Trillian'))
        self.assertSQL(
            "SELECT characters.name AS characters_name FROM characters " +
            "WHERE match(" +
            "(characters.info['race'] 0.9, characters.name 0.5), ?" +
            ")",
            query
        )

    def test_match_type_options(self):
        query = self.session.query(self.Character.name) \
            .filter(match({self.Character.name: 0.5,
                           self.Character.info['race']: 0.9},
                          'Trillian',
                          match_type='phrase',
                          options={'fuzziness': 3, 'analyzer': 'english'}))
        self.assertSQL(
            "SELECT characters.name AS characters_name FROM characters " +
            "WHERE match(" +
            "(characters.info['race'] 0.9, characters.name 0.5), ?" +
            ") using phrase with (analyzer=english, fuzziness=3)",
            query
        )

    def test_score(self):
        query = self.session.query(self.Character.name,
                                   sa.literal_column('_score')) \
                    .filter(match(self.Character.name, 'Trillian'))
        self.assertSQL(
            "SELECT characters.name AS characters_name, _score " +
            "FROM characters WHERE match(characters.name, ?)",
            query
        )

    def test_options_without_type(self):
        query = self.session.query(self.Character.name).filter(
            match({self.Character.name: 0.5, self.Character.info['race']: 0.9},
                  'Trillian',
                  options={'boost': 10.0})
        )
        err = None
        try:
            str(query)
        except ValueError as e:
            err = e
        msg = "missing match_type. " + \
              "It's not allowed to specify options without match_type"
        self.assertEqual(str(err), msg)



# -*- coding: utf-8; -*-
#
# Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
# license agreements.  See the NOTICE file distributed with this work for
# additional information regarding copyright ownership.  Crate licenses
# this file to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may
# obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
# However, if you have executed another commercial license agreement
# with Crate these terms will supersede the license and you may use the
# software solely pursuant to the terms of the relevant commercial agreement.

import hashlib


class BlobContainer(object):
    """ class that represents a blob collection in crate.

    can be used to download, upload and delete blobs
    """

    def __init__(self, container_name, connection):
        self.container_name = container_name
        self.conn = connection

    def _compute_digest(self, f):
        f.seek(0)
        m = hashlib.sha1()
        while True:
            d = f.read(1024 * 32)
            if not d:
                break
            m.update(d)
        f.seek(0)
        return m.hexdigest()

    def put(self, f, digest=None):
        """
        Upload a blob

        :param f:
            File object to be uploaded (required to support seek if digest is
            not provided).
        :param digest:
            Optional SHA-1 hex digest of the file contents. Gets computed
            before actual upload if not provided, which requires an extra file
            read.
        :return:
            The hex digest of the uploaded blob if not provided in the call.
            Otherwise a boolean indicating if the blob has been newly created.
        """

        if digest:
            actual_digest = digest
        else:
            actual_digest = self._compute_digest(f)

        created = self.conn.client.blob_put(self.container_name,
                                            actual_digest, f)
        if digest:
            return created
        return actual_digest

    def get(self, digest, chunk_size=1024 * 128):
        """
        Return the contents of a blob

        :param digest: the hex digest of the blob to return
        :param chunk_size: the size of the chunks returned on each iteration
        :return: generator returning chunks of data
        """
        return self.conn.client.blob_get(self.container_name, digest,
                                         chunk_size)

    def delete(self, digest):
        """
        Delete a blob

        :param digest: the hex digest of the blob to be deleted
        :return: True if blob existed
        """
        return self.conn.client.blob_del(self.container_name, digest)

    def exists(self, digest):
        """
        Check if a blob exists

        :param digest: Hex digest of the blob
        :return: Boolean indicating existence of the blob
        """
        return self.conn.client.blob_exists(self.container_name, digest)

    def __repr__(self):
        return "<BlobContainer '{0}'>".format(self.container_name)



# -*- coding: utf-8; -*-
#
# Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
# license agreements.  See the NOTICE file distributed with this work for
# additional information regarding copyright ownership.  Crate licenses
# this file to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may
# obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
# However, if you have executed another commercial license agreement
# with Crate these terms will supersede the license and you may use the
# software solely pursuant to the terms of the relevant commercial agreement.
import logging


logger = logging.getLogger(__name__)


def insert_bulk(pd_table, conn, keys, data_iter):
    """
    Use CrateDB's "bulk operations" endpoint as a fast path for pandas' and Dask's `to_sql()` [1] method.

    The idea is to break out of SQLAlchemy, compile the insert statement, and use the raw
    DBAPI connection client, in order to invoke a request using `bulk_parameters` [2]::

        cursor.execute(sql=sql, bulk_parameters=data)

    The vanilla implementation, used by SQLAlchemy, is::

        data = [dict(zip(keys, row)) for row in data_iter]
        conn.execute(pd_table.table.insert(), data)

    Batch chunking will happen outside of this function, for example [3] demonstrates
    the relevant code in `pandas.io.sql`.

    [1] https://pandas.pydata.org/docs/reference/api/pandas.DataFrame.to_sql.html
    [2] https://crate.io/docs/crate/reference/en/latest/interfaces/http.html#bulk-operations
    [3] https://github.com/pandas-dev/pandas/blob/v2.0.1/pandas/io/sql.py#L1011-L1027
    """

    # Compile SQL statement and materialize batch.
    sql = str(pd_table.table.insert().compile(bind=conn))
    data = list(data_iter)

    # For debugging and tracing the batches running through this method.
    if logger.level == logging.DEBUG:
        logger.debug(f"Bulk SQL:     {sql}")
        logger.debug(f"Bulk records: {len(data)}")
        # logger.debug(f"Bulk data:    {data}")

    # Invoke bulk insert operation.
    cursor = conn._dbapi_connection.cursor()
    cursor.execute(sql=sql, bulk_parameters=data)
    cursor.close()



# -*- coding: utf-8; -*-
#
# Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
# license agreements.  See the NOTICE file distributed with this work for
# additional information regarding copyright ownership.  Crate licenses
# this file to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may
# obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
# However, if you have executed another commercial license agreement
# with Crate these terms will supersede the license and you may use the
# software solely pursuant to the terms of the relevant commercial agreement.

import sqlalchemy as sa
try:
    from sqlalchemy.orm import declarative_base
except ImportError:
    from sqlalchemy.ext.declarative import declarative_base

from crate.client.sqlalchemy.types import ObjectType, ObjectArray, Geopoint
from crate.client.cursor import Cursor

from unittest import TestCase
from unittest.mock import patch, MagicMock


fake_cursor = MagicMock(name='fake_cursor')
FakeCursor = MagicMock(name='FakeCursor', spec=Cursor)
FakeCursor.return_value = fake_cursor


@patch('crate.client.connection.Cursor', FakeCursor)
class SqlAlchemyCreateTableTest(TestCase):

    def setUp(self):
        self.engine = sa.create_engine('crate://')
        self.Base = declarative_base()

    def test_table_basic_types(self):
        class User(self.Base):
            __tablename__ = 'users'
            string_col = sa.Column(sa.String, primary_key=True)
            unicode_col = sa.Column(sa.Unicode)
            text_col = sa.Column(sa.Text)
            int_col = sa.Column(sa.Integer)
            long_col1 = sa.Column(sa.BigInteger)
            long_col2 = sa.Column(sa.NUMERIC)
            bool_col = sa.Column(sa.Boolean)
            short_col = sa.Column(sa.SmallInteger)
            datetime_col = sa.Column(sa.DateTime)
            date_col = sa.Column(sa.Date)
            float_col = sa.Column(sa.Float)
            double_col = sa.Column(sa.DECIMAL)

        self.Base.metadata.create_all(bind=self.engine)
        fake_cursor.execute.assert_called_with(
            ('\nCREATE TABLE users (\n\tstring_col STRING NOT NULL, '
             '\n\tunicode_col STRING, \n\ttext_col STRING, \n\tint_col INT, '
             '\n\tlong_col1 LONG, \n\tlong_col2 LONG, '
             '\n\tbool_col BOOLEAN, '
             '\n\tshort_col SHORT, '
             '\n\tdatetime_col TIMESTAMP, \n\tdate_col TIMESTAMP, '
             '\n\tfloat_col FLOAT, \n\tdouble_col DOUBLE, '
             '\n\tPRIMARY KEY (string_col)\n)\n\n'),
            ())

    def test_column_obj(self):
        class DummyTable(self.Base):
            __tablename__ = 'dummy'
            pk = sa.Column(sa.String, primary_key=True)
            obj_col = sa.Column(ObjectType)
        self.Base.metadata.create_all(bind=self.engine)
        fake_cursor.execute.assert_called_with(
            ('\nCREATE TABLE dummy (\n\tpk STRING NOT NULL, \n\tobj_col OBJECT, '
             '\n\tPRIMARY KEY (pk)\n)\n\n'),
            ())

    def test_table_clustered_by(self):
        class DummyTable(self.Base):
            __tablename__ = 't'
            __table_args__ = {
                'crate_clustered_by': 'p'
            }
            pk = sa.Column(sa.String, primary_key=True)
            p = sa.Column(sa.String)
        self.Base.metadata.create_all(bind=self.engine)
        fake_cursor.execute.assert_called_with(
            ('\nCREATE TABLE t (\n\t'
             'pk STRING NOT NULL, \n\t'
             'p STRING, \n\t'
             'PRIMARY KEY (pk)\n'
             ') CLUSTERED BY (p)\n\n'),
            ())

    def test_column_computed(self):
        class DummyTable(self.Base):
            __tablename__ = 't'
            ts = sa.Column(sa.BigInteger, primary_key=True)
            p = sa.Column(sa.BigInteger, sa.Computed("date_trunc('day', ts)"))
        self.Base.metadata.create_all(bind=self.engine)
        fake_cursor.execute.assert_called_with(
            ('\nCREATE TABLE t (\n\t'
             'ts LONG NOT NULL, \n\t'
             'p LONG GENERATED ALWAYS AS (date_trunc(\'day\', ts)), \n\t'
             'PRIMARY KEY (ts)\n'
             ')\n\n'),
            ())

    def test_column_computed_virtual(self):
        class DummyTable(self.Base):
            __tablename__ = 't'
            ts = sa.Column(sa.BigInteger, primary_key=True)
            p = sa.Column(sa.BigInteger, sa.Computed("date_trunc('day', ts)", persisted=False))
        with self.assertRaises(sa.exc.CompileError):
            self.Base.metadata.create_all(bind=self.engine)

    def test_table_partitioned_by(self):
        class DummyTable(self.Base):
            __tablename__ = 't'
            __table_args__ = {
                'crate_partitioned_by': 'p',
                'invalid_option': 1
            }
            pk = sa.Column(sa.String, primary_key=True)
            p = sa.Column(sa.String)
        self.Base.metadata.create_all(bind=self.engine)
        fake_cursor.execute.assert_called_with(
            ('\nCREATE TABLE t (\n\t'
             'pk STRING NOT NULL, \n\t'
             'p STRING, \n\t'
             'PRIMARY KEY (pk)\n'
             ') PARTITIONED BY (p)\n\n'),
            ())

    def test_table_number_of_shards_and_replicas(self):
        class DummyTable(self.Base):
            __tablename__ = 't'
            __table_args__ = {
                'crate_number_of_replicas': '2',
                'crate_number_of_shards': 3
            }
            pk = sa.Column(sa.String, primary_key=True)

        self.Base.metadata.create_all(bind=self.engine)
        fake_cursor.execute.assert_called_with(
            ('\nCREATE TABLE t (\n\t'
             'pk STRING NOT NULL, \n\t'
             'PRIMARY KEY (pk)\n'
             ') CLUSTERED INTO 3 SHARDS WITH (NUMBER_OF_REPLICAS = 2)\n\n'),
            ())

    def test_table_clustered_by_and_number_of_shards(self):
        class DummyTable(self.Base):
            __tablename__ = 't'
            __table_args__ = {
                'crate_clustered_by': 'p',
                'crate_number_of_shards': 3
            }
            pk = sa.Column(sa.String, primary_key=True)
            p = sa.Column(sa.String, primary_key=True)
        self.Base.metadata.create_all(bind=self.engine)
        fake_cursor.execute.assert_called_with(
            ('\nCREATE TABLE t (\n\t'
             'pk STRING NOT NULL, \n\t'
             'p STRING NOT NULL, \n\t'
             'PRIMARY KEY (pk, p)\n'
             ') CLUSTERED BY (p) INTO 3 SHARDS\n\n'),
            ())

    def test_column_object_array(self):
        class DummyTable(self.Base):
            __tablename__ = 't'
            pk = sa.Column(sa.String, primary_key=True)
            tags = sa.Column(ObjectArray)

        self.Base.metadata.create_all(bind=self.engine)
        fake_cursor.execute.assert_called_with(
            ('\nCREATE TABLE t (\n\t'
             'pk STRING NOT NULL, \n\t'
             'tags ARRAY(OBJECT), \n\t'
             'PRIMARY KEY (pk)\n)\n\n'), ())

    def test_column_nullable(self):
        class DummyTable(self.Base):
            __tablename__ = 't'
            pk = sa.Column(sa.String, primary_key=True)
            a = sa.Column(sa.Integer, nullable=True)
            b = sa.Column(sa.Integer, nullable=False)

        self.Base.metadata.create_all(bind=self.engine)
        fake_cursor.execute.assert_called_with(
            ('\nCREATE TABLE t (\n\t'
             'pk STRING NOT NULL, \n\t'
             'a INT, \n\t'
             'b INT NOT NULL, \n\t'
             'PRIMARY KEY (pk)\n)\n\n'), ())

    def test_column_pk_nullable(self):
        class DummyTable(self.Base):
            __tablename__ = 't'
            pk = sa.Column(sa.String, primary_key=True, nullable=True)
        with self.assertRaises(sa.exc.CompileError):
            self.Base.metadata.create_all(bind=self.engine)

    def test_column_crate_index(self):
        class DummyTable(self.Base):
            __tablename__ = 't'
            pk = sa.Column(sa.String, primary_key=True)
            a = sa.Column(sa.Integer, crate_index=False)
            b = sa.Column(sa.Integer, crate_index=True)

        self.Base.metadata.create_all(bind=self.engine)
        fake_cursor.execute.assert_called_with(
            ('\nCREATE TABLE t (\n\t'
             'pk STRING NOT NULL, \n\t'
             'a INT INDEX OFF, \n\t'
             'b INT, \n\t'
             'PRIMARY KEY (pk)\n)\n\n'), ())

    def test_column_geopoint_without_index(self):
        class DummyTable(self.Base):
            __tablename__ = 't'
            pk = sa.Column(sa.String, primary_key=True)
            a = sa.Column(Geopoint, crate_index=False)
        with self.assertRaises(sa.exc.CompileError):
            self.Base.metadata.create_all(bind=self.engine)

    def test_text_column_without_columnstore(self):
        class DummyTable(self.Base):
            __tablename__ = 't'
            pk = sa.Column(sa.String, primary_key=True)
            a = sa.Column(sa.String, crate_columnstore=False)
            b = sa.Column(sa.String, crate_columnstore=True)
            c = sa.Column(sa.String)

        self.Base.metadata.create_all(bind=self.engine)

        fake_cursor.execute.assert_called_with(
            ('\nCREATE TABLE t (\n\t'
             'pk STRING NOT NULL, \n\t'
             'a STRING STORAGE WITH (columnstore = false), \n\t'
             'b STRING, \n\t'
             'c STRING, \n\t'
             'PRIMARY KEY (pk)\n)\n\n'), ())

    def test_non_text_column_without_columnstore(self):
        class DummyTable(self.Base):
            __tablename__ = 't'
            pk = sa.Column(sa.String, primary_key=True)
            a = sa.Column(sa.Integer, crate_columnstore=False)

        with self.assertRaises(sa.exc.CompileError):
            self.Base.metadata.create_all(bind=self.engine)

    def test_column_server_default_text_func(self):
        class DummyTable(self.Base):
            __tablename__ = 't'
            pk = sa.Column(sa.String, primary_key=True)
            a = sa.Column(sa.DateTime, server_default=sa.text("now()"))

        self.Base.metadata.create_all(bind=self.engine)
        fake_cursor.execute.assert_called_with(
            ('\nCREATE TABLE t (\n\t'
             'pk STRING NOT NULL, \n\t'
             'a TIMESTAMP DEFAULT now(), \n\t'
             'PRIMARY KEY (pk)\n)\n\n'), ())

    def test_column_server_default_string(self):
        class DummyTable(self.Base):
            __tablename__ = 't'
            pk = sa.Column(sa.String, primary_key=True)
            a = sa.Column(sa.String, server_default="Zaphod")

        self.Base.metadata.create_all(bind=self.engine)
        fake_cursor.execute.assert_called_with(
            ('\nCREATE TABLE t (\n\t'
             'pk STRING NOT NULL, \n\t'
             'a STRING DEFAULT \'Zaphod\', \n\t'
             'PRIMARY KEY (pk)\n)\n\n'), ())

    def test_column_server_default_func(self):
        class DummyTable(self.Base):
            __tablename__ = 't'
            pk = sa.Column(sa.String, primary_key=True)
            a = sa.Column(sa.DateTime, server_default=sa.func.now())

        self.Base.metadata.create_all(bind=self.engine)
        fake_cursor.execute.assert_called_with(
            ('\nCREATE TABLE t (\n\t'
             'pk STRING NOT NULL, \n\t'
             'a TIMESTAMP DEFAULT now(), \n\t'
             'PRIMARY KEY (pk)\n)\n\n'), ())

    def test_column_server_default_text_constant(self):
        class DummyTable(self.Base):
            __tablename__ = 't'
            pk = sa.Column(sa.String, primary_key=True)
            answer = sa.Column(sa.Integer, server_default=sa.text("42"))

        self.Base.metadata.create_all(bind=self.engine)
        fake_cursor.execute.assert_called_with(
            ('\nCREATE TABLE t (\n\t'
             'pk STRING NOT NULL, \n\t'
             'answer INT DEFAULT 42, \n\t'
             'PRIMARY KEY (pk)\n)\n\n'), ())



# -*- coding: utf-8; -*-
#
# Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
# license agreements.  See the NOTICE file distributed with this work for
# additional information regarding copyright ownership.  Crate licenses
# this file to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may
# obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
# However, if you have executed another commercial license agreement
# with Crate these terms will supersede the license and you may use the
# software solely pursuant to the terms of the relevant commercial agreement.

import string
import warnings
from collections import defaultdict

import sqlalchemy as sa
from sqlalchemy.dialects.postgresql.base import PGCompiler
from sqlalchemy.sql import compiler
from sqlalchemy.types import String
from .types import MutableDict, ObjectTypeImpl, Geopoint, Geoshape
from .sa_version import SA_VERSION, SA_1_4


def rewrite_update(clauseelement, multiparams, params):
    """ change the params to enable partial updates

    sqlalchemy by default only supports updates of complex types in the form of

        "col = ?", ({"x": 1, "y": 2}

    but crate supports

        "col['x'] = ?, col['y'] = ?", (1, 2)

    by using the `ObjectType` (`MutableDict`) type.
    The update statement is only rewritten if an item of the MutableDict was
    changed.
    """
    newmultiparams = []
    _multiparams = multiparams[0]
    if len(_multiparams) == 0:
        return clauseelement, multiparams, params
    for _params in _multiparams:
        newparams = {}
        for key, val in _params.items():
            if (
                not isinstance(val, MutableDict) or
                (not any(val._changed_keys) and not any(val._deleted_keys))
            ):
                newparams[key] = val
                continue

            for subkey, subval in val.items():
                if subkey in val._changed_keys:
                    newparams["{0}['{1}']".format(key, subkey)] = subval
            for subkey in val._deleted_keys:
                newparams["{0}['{1}']".format(key, subkey)] = None
        newmultiparams.append(newparams)
    _multiparams = (newmultiparams, )
    clause = clauseelement.values(newmultiparams[0])
    clause._crate_specific = True
    return clause, _multiparams, params


@sa.event.listens_for(sa.engine.Engine, "before_execute", retval=True)
def crate_before_execute(conn, clauseelement, multiparams, params, *args, **kwargs):
    is_crate = type(conn.dialect).__name__ == 'CrateDialect'
    if is_crate and isinstance(clauseelement, sa.sql.expression.Update):
        if SA_VERSION >= SA_1_4:
            if params is None:
                multiparams = ([],)
            else:
                multiparams = ([params],)
            params = {}

        clauseelement, multiparams, params = rewrite_update(clauseelement, multiparams, params)

        if SA_VERSION >= SA_1_4:
            if multiparams[0]:
                params = multiparams[0][0]
            else:
                params = multiparams[0]
            multiparams = []

    return clauseelement, multiparams, params


class CrateDDLCompiler(compiler.DDLCompiler):

    __special_opts_tmpl = {
        'PARTITIONED_BY': ' PARTITIONED BY ({0})'
    }
    __clustered_opts_tmpl = {
        'NUMBER_OF_SHARDS': ' INTO {0} SHARDS',
        'CLUSTERED_BY': ' BY ({0})',
    }
    __clustered_opt_tmpl = ' CLUSTERED{CLUSTERED_BY}{NUMBER_OF_SHARDS}'

    def get_column_specification(self, column, **kwargs):
        colspec = self.preparer.format_column(column) + " " + \
            self.dialect.type_compiler.process(column.type)

        default = self.get_column_default_string(column)
        if default is not None:
            colspec += " DEFAULT " + default

        if column.computed is not None:
            colspec += " " + self.process(column.computed)

        if column.nullable is False:
            colspec += " NOT NULL"
        elif column.nullable and column.primary_key:
            raise sa.exc.CompileError(
                "Primary key columns cannot be nullable"
            )

        if column.dialect_options['crate'].get('index') is False:
            if isinstance(column.type, (Geopoint, Geoshape, ObjectTypeImpl)):
                raise sa.exc.CompileError(
                    "Disabling indexing is not supported for column "
                    "types OBJECT, GEO_POINT, and GEO_SHAPE"
                )

            colspec += " INDEX OFF"

        if column.dialect_options['crate'].get('columnstore') is False:
            if not isinstance(column.type, (String, )):
                raise sa.exc.CompileError(
                    "Controlling the columnstore is only allowed for STRING columns"
                )

            colspec += " STORAGE WITH (columnstore = false)"

        return colspec

    def visit_computed_column(self, generated):
        if generated.persisted is False:
            raise sa.exc.CompileError(
                "Virtual computed columns are not supported, set "
                "'persisted' to None or True"
            )

        return "GENERATED ALWAYS AS (%s)" % self.sql_compiler.process(
            generated.sqltext, include_table=False, literal_binds=True
        )

    def post_create_table(self, table):
        special_options = ''
        clustered_options = defaultdict(str)
        table_opts = []

        opts = dict(
            (k[len(self.dialect.name) + 1:].upper(), v)
            for k, v, in table.kwargs.items()
            if k.startswith('%s_' % self.dialect.name)
        )
        for k, v in opts.items():
            if k in self.__special_opts_tmpl:
                special_options += self.__special_opts_tmpl[k].format(v)
            elif k in self.__clustered_opts_tmpl:
                clustered_options[k] = self.__clustered_opts_tmpl[k].format(v)
            else:
                table_opts.append('{0} = {1}'.format(k, v))
        if clustered_options:
            special_options += string.Formatter().vformat(
                self.__clustered_opt_tmpl, (), clustered_options)
        if table_opts:
            return special_options + ' WITH ({0})'.format(
                ', '.join(sorted(table_opts)))
        return special_options

    def visit_foreign_key_constraint(self, constraint, **kw):
        """
        CrateDB does not support foreign key constraints.
        """
        warnings.warn("CrateDB does not support foreign key constraints, "
                      "they will be omitted when generating DDL statements.")
        return None

    def visit_unique_constraint(self, constraint, **kw):
        """
        CrateDB does not support unique key constraints.
        """
        warnings.warn("CrateDB does not support unique constraints, "
                      "they will be omitted when generating DDL statements.")
        return None


class CrateTypeCompiler(compiler.GenericTypeCompiler):

    def visit_string(self, type_, **kw):
        return 'STRING'

    def visit_unicode(self, type_, **kw):
        return 'STRING'

    def visit_TEXT(self, type_, **kw):
        return 'STRING'

    def visit_DECIMAL(self, type_, **kw):
        return 'DOUBLE'

    def visit_BIGINT(self, type_, **kw):
        return 'LONG'

    def visit_NUMERIC(self, type_, **kw):
        return 'LONG'

    def visit_INTEGER(self, type_, **kw):
        return 'INT'

    def visit_SMALLINT(self, type_, **kw):
        return 'SHORT'

    def visit_datetime(self, type_, **kw):
        return 'TIMESTAMP'

    def visit_date(self, type_, **kw):
        return 'TIMESTAMP'

    def visit_ARRAY(self, type_, **kw):
        if type_.dimensions is not None and type_.dimensions > 1:
            raise NotImplementedError(
                "CrateDB doesn't support multidimensional arrays")
        return 'ARRAY({0})'.format(self.process(type_.item_type))

    def visit_OBJECT(self, type_, **kw):
        return "OBJECT"


class CrateCompiler(compiler.SQLCompiler):

    def visit_getitem_binary(self, binary, operator, **kw):
        return "{0}['{1}']".format(
            self.process(binary.left, **kw),
            binary.right.value
        )

    def visit_json_getitem_op_binary(
        self, binary, operator, _cast_applied=False, **kw
    ):
        return "{0}['{1}']".format(
            self.process(binary.left, **kw),
            binary.right.value
        )

    def visit_any(self, element, **kw):
        return "%s%sANY (%s)" % (
            self.process(element.left, **kw),
            compiler.OPERATORS[element.operator],
            self.process(element.right, **kw)
        )

    def visit_ilike_case_insensitive_operand(self, element, **kw):
        """
        Use native `ILIKE` operator, like PostgreSQL's `PGCompiler`.
        """
        if self.dialect.has_ilike_operator():
            return element.element._compiler_dispatch(self, **kw)
        else:
            return super().visit_ilike_case_insensitive_operand(element, **kw)

    def visit_ilike_op_binary(self, binary, operator, **kw):
        """
        Use native `ILIKE` operator, like PostgreSQL's `PGCompiler`.

        Do not implement the `ESCAPE` functionality, because it is not
        supported by CrateDB.
        """
        if binary.modifiers.get("escape", None) is not None:
            raise NotImplementedError("Unsupported feature: ESCAPE is not supported")
        if self.dialect.has_ilike_operator():
            return "%s ILIKE %s" % (
                self.process(binary.left, **kw),
                self.process(binary.right, **kw),
            )
        else:
            return super().visit_ilike_op_binary(binary, operator, **kw)

    def visit_not_ilike_op_binary(self, binary, operator, **kw):
        """
        Use native `ILIKE` operator, like PostgreSQL's `PGCompiler`.

        Do not implement the `ESCAPE` functionality, because it is not
        supported by CrateDB.
        """
        if binary.modifiers.get("escape", None) is not None:
            raise NotImplementedError("Unsupported feature: ESCAPE is not supported")
        if self.dialect.has_ilike_operator():
            return "%s NOT ILIKE %s" % (
                self.process(binary.left, **kw),
                self.process(binary.right, **kw),
            )
        else:
            return super().visit_not_ilike_op_binary(binary, operator, **kw)

    def limit_clause(self, select, **kw):
        """
        Generate OFFSET / LIMIT clause, PostgreSQL-compatible.
        """
        return PGCompiler.limit_clause(self, select, **kw)

    def for_update_clause(self, select, **kw):
        # CrateDB does not support the `INSERT ... FOR UPDATE` clause.
        # See https://github.com/crate/crate-python/issues/577.
        warnings.warn("CrateDB does not support the 'INSERT ... FOR UPDATE' clause, "
                      "it will be omitted when generating SQL statements.")
        return ''



# vi: set encoding=utf-8
# -*- coding: utf-8; -*-
#
# Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
# license agreements.  See the NOTICE file distributed with this work for
# additional information regarding copyright ownership.  Crate licenses
# this file to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may
# obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
# However, if you have executed another commercial license agreement
# with Crate these terms will supersede the license and you may use the
# software solely pursuant to the terms of the relevant commercial agreement.
from __future__ import absolute_import

import os


def docs_path(*parts):
    return os.path.abspath(
        os.path.join(
            os.path.dirname(os.path.dirname(__file__)), *parts
        )
    )


def project_root(*parts):
    return os.path.abspath(
        os.path.join(docs_path("..", ".."), *parts)
    )


def crate_path(*parts):
    return os.path.abspath(
        project_root("parts", "crate", *parts)
    )


crate_port = 44209
crate_transport_port = 44309
localhost = '127.0.0.1'
crate_host = "{host}:{port}".format(host=localhost, port=crate_port)
crate_uri = "http://%s" % crate_host



# -*- coding: utf-8; -*-
#
# Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
# license agreements.  See the NOTICE file distributed with this work for
# additional information regarding copyright ownership.  Crate licenses
# this file to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may
# obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
# However, if you have executed another commercial license agreement
# with Crate these terms will supersede the license and you may use the
# software solely pursuant to the terms of the relevant commercial agreement.
from datetime import datetime, timedelta, timezone

from .converter import DataType
import warnings
import typing as t

from .converter import Converter
from .exceptions import ProgrammingError


class Cursor(object):
    """
    not thread-safe by intention
    should not be shared between different threads
    """
    lastrowid = None  # currently not supported

    def __init__(self, connection, converter: Converter, **kwargs):
        self.arraysize = 1
        self.connection = connection
        self._converter = converter
        self._closed = False
        self._result = None
        self.rows = None
        self._time_zone = None
        self.time_zone = kwargs.get("time_zone")

    def execute(self, sql, parameters=None, bulk_parameters=None):
        """
        Prepare and execute a database operation (query or command).
        """
        if self.connection._closed:
            raise ProgrammingError("Connection closed")

        if self._closed:
            raise ProgrammingError("Cursor closed")

        self._result = self.connection.client.sql(sql, parameters,
                                                  bulk_parameters)
        if "rows" in self._result:
            if self._converter is None:
                self.rows = iter(self._result["rows"])
            else:
                self.rows = iter(self._convert_rows())

    def executemany(self, sql, seq_of_parameters):
        """
        Prepare a database operation (query or command) and then execute it
        against all parameter sequences or mappings found in the sequence
        ``seq_of_parameters``.
        """
        row_counts = []
        durations = []
        self.execute(sql, bulk_parameters=seq_of_parameters)

        for result in self._result.get('results', []):
            if result.get('rowcount') > -1:
                row_counts.append(result.get('rowcount'))
        if self.duration > -1:
            durations.append(self.duration)

        self._result = {
            "rowcount": sum(row_counts) if row_counts else -1,
            "duration": sum(durations) if durations else -1,
            "rows": [],
            "cols": self._result.get("cols", []),
            "col_types": self._result.get("col_types", []),
            "results": self._result.get("results")
        }
        if self._converter is None:
            self.rows = iter(self._result["rows"])
        else:
            self.rows = iter(self._convert_rows())
        return self._result["results"]

    def fetchone(self):
        """
        Fetch the next row of a query result set, returning a single sequence,
        or None when no more data is available.
        Alias for ``next()``.
        """
        try:
            return self.next()
        except StopIteration:
            return None

    def __iter__(self):
        """
        support iterator interface:
        http://legacy.python.org/dev/peps/pep-0249/#iter

        This iterator is shared. Advancing this iterator will advance other
        iterators created from this cursor.
        """
        warnings.warn("DB-API extension cursor.__iter__() used")
        return self

    def fetchmany(self, count=None):
        """
        Fetch the next set of rows of a query result, returning a sequence of
        sequences (e.g. a list of tuples). An empty sequence is returned when
        no more rows are available.
        """
        if count is None:
            count = self.arraysize
        if count == 0:
            return self.fetchall()
        result = []
        for i in range(count):
            try:
                result.append(self.next())
            except StopIteration:
                pass
        return result

    def fetchall(self):
        """
        Fetch all (remaining) rows of a query result, returning them as a
        sequence of sequences (e.g. a list of tuples). Note that the cursor's
        arraysize attribute can affect the performance of this operation.
        """
        result = []
        iterate = True
        while iterate:
            try:
                result.append(self.next())
            except StopIteration:
                iterate = False
        return result

    def close(self):
        """
        Close the cursor now
        """
        self._closed = True
        self._result = None

    def setinputsizes(self, sizes):
        """
        Not supported method.
        """
        pass

    def setoutputsize(self, size, column=None):
        """
        Not supported method.
        """
        pass

    @property
    def rowcount(self):
        """
        This read-only attribute specifies the number of rows that the last
        .execute*() produced (for DQL statements like ``SELECT``) or affected
        (for DML statements like ``UPDATE`` or ``INSERT``).
        """
        if (self._closed or not self._result or "rows" not in self._result):
            return -1
        return self._result.get("rowcount", -1)

    def next(self):
        """
        Return the next row of a query result set, respecting if cursor was
        closed.
        """
        if self.rows is None:
            raise ProgrammingError(
                "No result available. " +
                "execute() or executemany() must be called first."
            )
        elif not self._closed:
            return next(self.rows)
        else:
            raise ProgrammingError("Cursor closed")

    __next__ = next

    @property
    def description(self):
        """
        This read-only attribute is a sequence of 7-item sequences.
        """
        if self._closed:
            return

        description = []
        for col in self._result["cols"]:
            description.append((col,
                                None,
                                None,
                                None,
                                None,
                                None,
                                None))
        return tuple(description)

    @property
    def duration(self):
        """
        This read-only attribute specifies the server-side duration of a query
        in milliseconds.
        """
        if self._closed or \
                not self._result or \
                "duration" not in self._result:
            return -1
        return self._result.get("duration", 0)

    def _convert_rows(self):
        """
        Iterate rows, apply type converters, and generate converted rows.
        """
        assert "col_types" in self._result and self._result["col_types"], \
               "Unable to apply type conversion without `col_types` information"

        # Resolve `col_types` definition to converter functions. Running the lookup
        # redundantly on each row loop iteration would be a huge performance hog.
        types = self._result["col_types"]
        converters = [
            self._converter.get(type) for type in types
        ]

        # Process result rows with conversion.
        for row in self._result["rows"]:
            yield [
                convert(value)
                for convert, value in zip(converters, row)
            ]

    @property
    def time_zone(self):
        """
        Get the current time zone.
        """
        return self._time_zone

    @time_zone.setter
    def time_zone(self, tz):
        """
        Set the time zone.

        Different data types are supported. Available options are:

        - ``datetime.timezone.utc``
        - ``datetime.timezone(datetime.timedelta(hours=7), name="MST")``
        - ``pytz.timezone("Australia/Sydney")``
        - ``zoneinfo.ZoneInfo("Australia/Sydney")``
        - ``+0530`` (UTC offset in string format)

        When `time_zone` is `None`, the returned `datetime` objects are
        "naive", without any `tzinfo`, converted using ``datetime.utcfromtimestamp(...)``.

        When `time_zone` is given, the returned `datetime` objects are "aware",
        with `tzinfo` set, converted using ``datetime.fromtimestamp(..., tz=...)``.
        """

        # Do nothing when time zone is reset.
        if tz is None:
            self._time_zone = None
            return

        # Requesting datetime-aware `datetime` objects needs the data type converter.
        # Implicitly create one, when needed.
        if self._converter is None:
            self._converter = Converter()

        # When the time zone is given as a string, assume UTC offset format, e.g. `+0530`.
        if isinstance(tz, str):
            tz = self._timezone_from_utc_offset(tz)

        self._time_zone = tz

        def _to_datetime_with_tz(value: t.Optional[float]) -> t.Optional[datetime]:
            """
            Convert CrateDB's `TIMESTAMP` value to a native Python `datetime`
            object, with timezone-awareness.
            """
            if value is None:
                return None
            return datetime.fromtimestamp(value / 1e3, tz=self._time_zone)

        # Register converter function for `TIMESTAMP` type.
        self._converter.set(DataType.TIMESTAMP_WITH_TZ, _to_datetime_with_tz)
        self._converter.set(DataType.TIMESTAMP_WITHOUT_TZ, _to_datetime_with_tz)

    @staticmethod
    def _timezone_from_utc_offset(tz) -> timezone:
        """
        Convert UTC offset in string format (e.g. `+0530`) into `datetime.timezone` object.
        """
        assert len(tz) == 5, f"Time zone '{tz}' is given in invalid UTC offset format"
        try:
            hours = int(tz[:3])
            minutes = int(tz[0] + tz[3:])
            return timezone(timedelta(hours=hours, minutes=minutes), name=tz)
        except Exception as ex:
            raise ValueError(f"Time zone '{tz}' is given in invalid UTC offset format: {ex}")



# -*- coding: utf-8; -*-
#
# Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
# license agreements.  See the NOTICE file distributed with this work for
# additional information regarding copyright ownership.  Crate licenses
# this file to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may
# obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
# However, if you have executed another commercial license agreement
# with Crate these terms will supersede the license and you may use the
# software solely pursuant to the terms of the relevant commercial agreement.

from __future__ import absolute_import
from datetime import datetime, tzinfo, timedelta
from unittest import TestCase
from unittest.mock import patch, MagicMock

import sqlalchemy as sa
from sqlalchemy.exc import DBAPIError
from sqlalchemy.orm import Session
try:
    from sqlalchemy.orm import declarative_base
except ImportError:
    from sqlalchemy.ext.declarative import declarative_base

from crate.client.cursor import Cursor


fake_cursor = MagicMock(name='fake_cursor')
FakeCursor = MagicMock(name='FakeCursor', spec=Cursor)
FakeCursor.return_value = fake_cursor


class CST(tzinfo):
    """
    Timezone object for CST
    """

    def utcoffset(self, date_time):
        return timedelta(seconds=-3600)

    def dst(self, date_time):
        return timedelta(seconds=-7200)


@patch('crate.client.connection.Cursor', FakeCursor)
class SqlAlchemyDateAndDateTimeTest(TestCase):

    def setUp(self):
        self.engine = sa.create_engine('crate://')
        Base = declarative_base()

        class Character(Base):
            __tablename__ = 'characters'
            name = sa.Column(sa.String, primary_key=True)
            date = sa.Column(sa.Date)
            timestamp = sa.Column(sa.DateTime)

        fake_cursor.description = (
            ('characters_name', None, None, None, None, None, None),
            ('characters_date', None, None, None, None, None, None)
        )
        self.session = Session(bind=self.engine)
        self.Character = Character

    def test_date_can_handle_datetime(self):
        """ date type should also be able to handle iso datetime strings.

        this verifies that the fallback in the Date result_processor works.
        """
        fake_cursor.fetchall.return_value = [
            ('Trillian', '2013-07-16T00:00:00.000Z')
        ]
        self.session.query(self.Character).first()

    def test_date_cannot_handle_tz_aware_datetime(self):
        character = self.Character()
        character.name = "Athur"
        character.timestamp = datetime(2009, 5, 13, 19, 19, 30, tzinfo=CST())
        self.session.add(character)
        self.assertRaises(DBAPIError, self.session.commit)



# -*- coding: utf-8; -*-
#
# Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
# license agreements.  See the NOTICE file distributed with this work for
# additional information regarding copyright ownership.  Crate licenses
# this file to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may
# obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
# However, if you have executed another commercial license agreement
# with Crate these terms will supersede the license and you may use the
# software solely pursuant to the terms of the relevant commercial agreement.

from typing import Any, Dict, List, MutableMapping, Optional, Tuple, Union

import sqlalchemy as sa
from sqlalchemy import ColumnClause, ValuesBase, cast, exc
from sqlalchemy.sql import dml
from sqlalchemy.sql.base import _from_objects
from sqlalchemy.sql.compiler import SQLCompiler
from sqlalchemy.sql.crud import (REQUIRED, _as_dml_column, _create_bind_param,
                                 _CrudParamElement, _CrudParams,
                                 _extend_values_for_multiparams,
                                 _get_stmt_parameter_tuples_params,
                                 _get_update_multitable_params,
                                 _key_getters_for_crud_column, _scan_cols,
                                 _scan_insert_from_select_cols,
                                 _setup_delete_return_defaults)
from sqlalchemy.sql.dml import DMLState, _DMLColumnElement
from sqlalchemy.sql.dml import isinsert as _compile_state_isinsert

from crate.client.sqlalchemy.compiler import CrateCompiler


class CrateCompilerSA20(CrateCompiler):

    def visit_update(self, update_stmt, **kw):
        compile_state = update_stmt._compile_state_factory(
            update_stmt, self, **kw
        )
        update_stmt = compile_state.statement

        # [20] CrateDB patch.
        if not compile_state._dict_parameters and \
                not hasattr(update_stmt, '_crate_specific'):
            return super().visit_update(update_stmt, **kw)

        toplevel = not self.stack
        if toplevel:
            self.isupdate = True
            if not self.dml_compile_state:
                self.dml_compile_state = compile_state
            if not self.compile_state:
                self.compile_state = compile_state

        extra_froms = compile_state._extra_froms
        is_multitable = bool(extra_froms)

        if is_multitable:
            # main table might be a JOIN
            main_froms = set(_from_objects(update_stmt.table))
            render_extra_froms = [
                f for f in extra_froms if f not in main_froms
            ]
            correlate_froms = main_froms.union(extra_froms)
        else:
            render_extra_froms = []
            correlate_froms = {update_stmt.table}

        self.stack.append(
            {
                "correlate_froms": correlate_froms,
                "asfrom_froms": correlate_froms,
                "selectable": update_stmt,
            }
        )

        text = "UPDATE "

        if update_stmt._prefixes:
            text += self._generate_prefixes(
                update_stmt, update_stmt._prefixes, **kw
            )

        table_text = self.update_tables_clause(
            update_stmt, update_stmt.table, render_extra_froms, **kw
        )
        # [20] CrateDB patch.
        crud_params_struct = _get_crud_params(
            self, update_stmt, compile_state, toplevel, **kw
        )
        crud_params = crud_params_struct.single_params

        if update_stmt._hints:
            dialect_hints, table_text = self._setup_crud_hints(
                update_stmt, table_text
            )
        else:
            dialect_hints = None

        if update_stmt._independent_ctes:
            self._dispatch_independent_ctes(update_stmt, kw)

        text += table_text

        text += " SET "

        # [20] CrateDB patch begin.
        include_table = extra_froms and \
            self.render_table_with_column_in_update_from

        set_clauses = []

        for c, expr, value, _ in crud_params:
            key = c._compiler_dispatch(self, include_table=include_table)
            clause = key + ' = ' + value
            set_clauses.append(clause)

        for k, v in compile_state._dict_parameters.items():
            if isinstance(k, str) and '[' in k:
                bindparam = sa.sql.bindparam(k, v)
                clause = k + ' = ' + self.process(bindparam)
                set_clauses.append(clause)

        text += ', '.join(set_clauses)
        # [20] CrateDB patch end.

        if self.implicit_returning or update_stmt._returning:
            if self.returning_precedes_values:
                text += " " + self.returning_clause(
                    update_stmt,
                    self.implicit_returning or update_stmt._returning,
                    populate_result_map=toplevel,
                )

        if extra_froms:
            extra_from_text = self.update_from_clause(
                update_stmt,
                update_stmt.table,
                render_extra_froms,
                dialect_hints,
                **kw,
            )
            if extra_from_text:
                text += " " + extra_from_text

        if update_stmt._where_criteria:
            t = self._generate_delimited_and_list(
                update_stmt._where_criteria, **kw
            )
            if t:
                text += " WHERE " + t

        limit_clause = self.update_limit_clause(update_stmt)
        if limit_clause:
            text += " " + limit_clause

        if (
            self.implicit_returning or update_stmt._returning
        ) and not self.returning_precedes_values:
            text += " " + self.returning_clause(
                update_stmt,
                self.implicit_returning or update_stmt._returning,
                populate_result_map=toplevel,
            )

        if self.ctes:
            nesting_level = len(self.stack) if not toplevel else None
            text = self._render_cte_clause(nesting_level=nesting_level) + text

        self.stack.pop(-1)

        return text


def _get_crud_params(
    compiler: SQLCompiler,
    stmt: ValuesBase,
    compile_state: DMLState,
    toplevel: bool,
    **kw: Any,
) -> _CrudParams:
    """create a set of tuples representing column/string pairs for use
    in an INSERT or UPDATE statement.

    Also generates the Compiled object's postfetch, prefetch, and
    returning column collections, used for default handling and ultimately
    populating the CursorResult's prefetch_cols() and postfetch_cols()
    collections.

    """

    # note: the _get_crud_params() system was written with the notion in mind
    # that INSERT, UPDATE, DELETE are always the top level statement and
    # that there is only one of them.  With the addition of CTEs that can
    # make use of DML, this assumption is no longer accurate; the DML
    # statement is not necessarily the top-level "row returning" thing
    # and it is also theoretically possible (fortunately nobody has asked yet)
    # to have a single statement with multiple DMLs inside of it via CTEs.

    # the current _get_crud_params() design doesn't accommodate these cases
    # right now.  It "just works" for a CTE that has a single DML inside of
    # it, and for a CTE with multiple DML, it's not clear what would happen.

    # overall, the "compiler.XYZ" collections here would need to be in a
    # per-DML structure of some kind, and DefaultDialect would need to
    # navigate these collections on a per-statement basis, with additional
    # emphasis on the "toplevel returning data" statement.  However we
    # still need to run through _get_crud_params() for all DML as we have
    # Python / SQL generated column defaults that need to be rendered.

    # if there is user need for this kind of thing, it's likely a post 2.0
    # kind of change as it would require deep changes to DefaultDialect
    # as well as here.

    compiler.postfetch = []
    compiler.insert_prefetch = []
    compiler.update_prefetch = []
    compiler.implicit_returning = []

    # getters - these are normally just column.key,
    # but in the case of mysql multi-table update, the rules for
    # .key must conditionally take tablename into account
    (
        _column_as_key,
        _getattr_col_key,
        _col_bind_name,
    ) = _key_getters_for_crud_column(compiler, stmt, compile_state)

    compiler._get_bind_name_for_col = _col_bind_name

    if stmt._returning and stmt._return_defaults:
        raise exc.CompileError(
            "Can't compile statement that includes returning() and "
            "return_defaults() simultaneously"
        )

    if compile_state.isdelete:
        _setup_delete_return_defaults(
            compiler,
            stmt,
            compile_state,
            (),
            _getattr_col_key,
            _column_as_key,
            _col_bind_name,
            (),
            (),
            toplevel,
            kw,
        )
        return _CrudParams([], [])

    # no parameters in the statement, no parameters in the
    # compiled params - return binds for all columns
    if compiler.column_keys is None and compile_state._no_parameters:
        return _CrudParams(
            [
                (
                    c,
                    compiler.preparer.format_column(c),
                    _create_bind_param(compiler, c, None, required=True),
                    (c.key,),
                )
                for c in stmt.table.columns
            ],
            [],
        )

    stmt_parameter_tuples: Optional[
        List[Tuple[Union[str, ColumnClause[Any]], Any]]
    ]
    spd: Optional[MutableMapping[_DMLColumnElement, Any]]

    if (
        _compile_state_isinsert(compile_state)
        and compile_state._has_multi_parameters
    ):
        mp = compile_state._multi_parameters
        assert mp is not None
        spd = mp[0]
        stmt_parameter_tuples = list(spd.items())
    elif compile_state._ordered_values:
        spd = compile_state._dict_parameters
        stmt_parameter_tuples = compile_state._ordered_values
    elif compile_state._dict_parameters:
        spd = compile_state._dict_parameters
        stmt_parameter_tuples = list(spd.items())
    else:
        stmt_parameter_tuples = spd = None

    # if we have statement parameters - set defaults in the
    # compiled params
    if compiler.column_keys is None:
        parameters = {}
    elif stmt_parameter_tuples:
        assert spd is not None
        parameters = {
            _column_as_key(key): REQUIRED
            for key in compiler.column_keys
            if key not in spd
        }
    else:
        parameters = {
            _column_as_key(key): REQUIRED for key in compiler.column_keys
        }

    # create a list of column assignment clauses as tuples
    values: List[_CrudParamElement] = []

    if stmt_parameter_tuples is not None:
        _get_stmt_parameter_tuples_params(
            compiler,
            compile_state,
            parameters,
            stmt_parameter_tuples,
            _column_as_key,
            values,
            kw,
        )

    check_columns: Dict[str, ColumnClause[Any]] = {}

    # special logic that only occurs for multi-table UPDATE
    # statements
    if dml.isupdate(compile_state) and compile_state.is_multitable:
        _get_update_multitable_params(
            compiler,
            stmt,
            compile_state,
            stmt_parameter_tuples,
            check_columns,
            _col_bind_name,
            _getattr_col_key,
            values,
            kw,
        )

    if _compile_state_isinsert(compile_state) and stmt._select_names:
        # is an insert from select, is not a multiparams

        assert not compile_state._has_multi_parameters

        _scan_insert_from_select_cols(
            compiler,
            stmt,
            compile_state,
            parameters,
            _getattr_col_key,
            _column_as_key,
            _col_bind_name,
            check_columns,
            values,
            toplevel,
            kw,
        )
    else:
        _scan_cols(
            compiler,
            stmt,
            compile_state,
            parameters,
            _getattr_col_key,
            _column_as_key,
            _col_bind_name,
            check_columns,
            values,
            toplevel,
            kw,
        )

    # [20] CrateDB patch.
    #
    # This sanity check performed by SQLAlchemy currently needs to be
    # deactivated in order to satisfy the rewriting logic of the CrateDB
    # dialect in `rewrite_update` and `visit_update`.
    #
    # It can be quickly reproduced by activating this section and running the
    # test cases::
    #
    #   ./bin/test -vvvv -t dict_test
    #
    # That croaks like::
    #
    #   sqlalchemy.exc.CompileError: Unconsumed column names: characters_name
    #
    # TODO: Investigate why this is actually happening and eventually mitigate
    #       the root cause.
    """
    if parameters and stmt_parameter_tuples:
        check = (
            set(parameters)
            .intersection(_column_as_key(k) for k, v in stmt_parameter_tuples)
            .difference(check_columns)
        )
        if check:
            raise exc.CompileError(
                "Unconsumed column names: %s"
                % (", ".join("%s" % (c,) for c in check))
            )
    """

    if (
        _compile_state_isinsert(compile_state)
        and compile_state._has_multi_parameters
    ):
        # is a multiparams, is not an insert from a select
        assert not stmt._select_names
        multi_extended_values = _extend_values_for_multiparams(
            compiler,
            stmt,
            compile_state,
            cast(
                "Sequence[_CrudParamElementStr]",
                values,
            ),
            cast("Callable[..., str]", _column_as_key),
            kw,
        )
        return _CrudParams(values, multi_extended_values)
    elif (
        not values
        and compiler.for_executemany
        and compiler.dialect.supports_default_metavalue
    ):
        # convert an "INSERT DEFAULT VALUES"
        # into INSERT (firstcol) VALUES (DEFAULT) which can be turned
        # into an in-place multi values.  This supports
        # insert_executemany_returning mode :)
        values = [
            (
                _as_dml_column(stmt.table.columns[0]),
                compiler.preparer.format_column(stmt.table.columns[0]),
                compiler.dialect.default_metavalue_token,
                (),
            )
        ]

    return _CrudParams(values, [])



# -*- coding: utf-8; -*-
#
# Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
# license agreements.  See the NOTICE file distributed with this work for
# additional information regarding copyright ownership.  Crate licenses
# this file to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may
# obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
# However, if you have executed another commercial license agreement
# with Crate these terms will supersede the license and you may use the
# software solely pursuant to the terms of the relevant commercial agreement.

from .compat.api13 import monkeypatch_add_exec_driver_sql
from .dialect import CrateDialect
from .sa_version import SA_1_4, SA_VERSION


if SA_VERSION < SA_1_4:
    import textwrap
    import warnings

    # SQLAlchemy 1.3 is effectively EOL.
    SA13_DEPRECATION_WARNING = textwrap.dedent("""
    WARNING: SQLAlchemy 1.3 is effectively EOL.

    SQLAlchemy 1.3 is EOL since 2023-01-27.
    Future versions of the CrateDB SQLAlchemy dialect will drop support for SQLAlchemy 1.3.
    It is recommended that you transition to using SQLAlchemy 1.4 or 2.0:

    - https://docs.sqlalchemy.org/en/14/changelog/migration_14.html
    - https://docs.sqlalchemy.org/en/20/changelog/migration_20.html
    """.lstrip("\n"))
    warnings.warn(message=SA13_DEPRECATION_WARNING, category=DeprecationWarning)

    # SQLAlchemy 1.3 does not have the `exec_driver_sql` method, so add it.
    monkeypatch_add_exec_driver_sql()


__all__ = [
    CrateDialect,
]



# -*- coding: utf-8; -*-
#
# Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
# license agreements.  See the NOTICE file distributed with this work for
# additional information regarding copyright ownership.  Crate licenses
# this file to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may
# obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
# However, if you have executed another commercial license agreement
# with Crate these terms will supersede the license and you may use the
# software solely pursuant to the terms of the relevant commercial agreement.

from datetime import datetime
from unittest import TestCase, skipIf
from unittest.mock import MagicMock, patch

import sqlalchemy as sa

from crate.client.cursor import Cursor
from crate.client.sqlalchemy import SA_VERSION
from crate.client.sqlalchemy.sa_version import SA_1_4, SA_2_0
from crate.client.sqlalchemy.types import ObjectType
from sqlalchemy import inspect
from sqlalchemy.orm import Session
try:
    from sqlalchemy.orm import declarative_base
except ImportError:
    from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.testing import eq_, in_, is_true

FakeCursor = MagicMock(name='FakeCursor', spec=Cursor)


@patch('crate.client.connection.Cursor', FakeCursor)
class SqlAlchemyDialectTest(TestCase):

    def execute_wrapper(self, query, *args, **kwargs):
        self.executed_statement = query
        return self.fake_cursor

    def setUp(self):

        self.fake_cursor = MagicMock(name='fake_cursor')
        FakeCursor.return_value = self.fake_cursor

        self.engine = sa.create_engine('crate://')

        self.executed_statement = None

        self.connection = self.engine.connect()

        self.fake_cursor.execute = self.execute_wrapper

        self.base = declarative_base()

        class Character(self.base):
            __tablename__ = 'characters'

            name = sa.Column(sa.String, primary_key=True)
            age = sa.Column(sa.Integer, primary_key=True)
            obj = sa.Column(ObjectType)
            ts = sa.Column(sa.DateTime, onupdate=datetime.utcnow)

        self.session = Session(bind=self.engine)

    def init_mock(self, return_value=None):
        self.fake_cursor.rowcount = 1
        self.fake_cursor.description = (
            ('foo', None, None, None, None, None, None),
        )
        self.fake_cursor.fetchall = MagicMock(return_value=return_value)

    def test_primary_keys_2_3_0(self):
        insp = inspect(self.session.bind)
        self.engine.dialect.server_version_info = (2, 3, 0)

        self.fake_cursor.rowcount = 3
        self.fake_cursor.description = (
            ('foo', None, None, None, None, None, None),
        )
        self.fake_cursor.fetchall = MagicMock(return_value=[["id"], ["id2"], ["id3"]])

        eq_(insp.get_pk_constraint("characters")['constrained_columns'], {"id", "id2", "id3"})
        self.fake_cursor.fetchall.assert_called_once_with()
        in_("information_schema.key_column_usage", self.executed_statement)
        in_("table_catalog = ?", self.executed_statement)

    def test_primary_keys_3_0_0(self):
        insp = inspect(self.session.bind)
        self.engine.dialect.server_version_info = (3, 0, 0)

        self.fake_cursor.rowcount = 3
        self.fake_cursor.description = (
            ('foo', None, None, None, None, None, None),
        )
        self.fake_cursor.fetchall = MagicMock(return_value=[["id"], ["id2"], ["id3"]])

        eq_(insp.get_pk_constraint("characters")['constrained_columns'], {"id", "id2", "id3"})
        self.fake_cursor.fetchall.assert_called_once_with()
        in_("information_schema.key_column_usage", self.executed_statement)
        in_("table_schema = ?", self.executed_statement)

    def test_get_table_names(self):
        self.fake_cursor.rowcount = 1
        self.fake_cursor.description = (
            ('foo', None, None, None, None, None, None),
        )
        self.fake_cursor.fetchall = MagicMock(return_value=[["t1"], ["t2"]])

        insp = inspect(self.session.bind)
        self.engine.dialect.server_version_info = (2, 0, 0)
        eq_(insp.get_table_names(schema="doc"),
            ['t1', 't2'])
        in_("WHERE table_schema = ? AND table_type = 'BASE TABLE' ORDER BY", self.executed_statement)

    def test_get_view_names(self):
        self.fake_cursor.rowcount = 1
        self.fake_cursor.description = (
            ('foo', None, None, None, None, None, None),
        )
        self.fake_cursor.fetchall = MagicMock(return_value=[["v1"], ["v2"]])

        insp = inspect(self.session.bind)
        self.engine.dialect.server_version_info = (2, 0, 0)
        eq_(insp.get_view_names(schema="doc"),
            ['v1', 'v2'])
        eq_(self.executed_statement, "SELECT table_name FROM information_schema.views "
                                     "ORDER BY table_name ASC, table_schema ASC")

    @skipIf(SA_VERSION < SA_1_4, "Inspector.has_table only available on SQLAlchemy>=1.4")
    def test_has_table(self):
        self.init_mock(return_value=[["foo"], ["bar"]])
        insp = inspect(self.session.bind)
        is_true(insp.has_table("bar"))
        eq_(self.executed_statement,
            "SELECT table_name FROM information_schema.tables "
            "WHERE table_schema = ? AND table_type = 'BASE TABLE' "
            "ORDER BY table_name ASC, table_schema ASC")

    @skipIf(SA_VERSION < SA_2_0, "Inspector.has_schema only available on SQLAlchemy>=2.0")
    def test_has_schema(self):
        self.init_mock(
            return_value=[["blob"], ["doc"], ["information_schema"], ["pg_catalog"], ["sys"]])
        insp = inspect(self.session.bind)
        is_true(insp.has_schema("doc"))
        eq_(self.executed_statement,
            "select schema_name from information_schema.schemata order by schema_name asc")



# -*- coding: utf-8; -*-
#
# Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
# license agreements.  See the NOTICE file distributed with this work for
# additional information regarding copyright ownership.  Crate licenses
# this file to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may
# obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
# However, if you have executed another commercial license agreement
# with Crate these terms will supersede the license and you may use the
# software solely pursuant to the terms of the relevant commercial agreement.
import json
import os
import tempfile
import urllib
from verlib2 import Version
from unittest import TestCase, mock
from io import BytesIO

import urllib3

import crate
from .layer import CrateLayer, prepend_http, http_url_from_host_port, wait_for_http_url
from .settings import crate_path


class LayerUtilsTest(TestCase):

    def test_prepend_http(self):
        host = prepend_http('localhost')
        self.assertEqual('http://localhost', host)
        host = prepend_http('http://localhost')
        self.assertEqual('http://localhost', host)
        host = prepend_http('https://localhost')
        self.assertEqual('https://localhost', host)
        host = prepend_http('http')
        self.assertEqual('http://http', host)

    def test_http_url(self):
        url = http_url_from_host_port(None, None)
        self.assertEqual(None, url)
        url = http_url_from_host_port('localhost', None)
        self.assertEqual(None, url)
        url = http_url_from_host_port(None, 4200)
        self.assertEqual(None, url)
        url = http_url_from_host_port('localhost', 4200)
        self.assertEqual('http://localhost:4200', url)
        url = http_url_from_host_port('https://crate', 4200)
        self.assertEqual('https://crate:4200', url)

    def test_wait_for_http(self):
        log = BytesIO(b'[i.c.p.h.CrateNettyHttpServerTransport] [crate] publish_address {127.0.0.1:4200}')
        addr = wait_for_http_url(log)
        self.assertEqual('http://127.0.0.1:4200', addr)
        log = BytesIO(b'[i.c.p.h.CrateNettyHttpServerTransport] [crate] publish_address {}')
        addr = wait_for_http_url(log=log, timeout=1)
        self.assertEqual(None, addr)

    @mock.patch.object(crate.testing.layer, "_download_and_extract", lambda uri, directory: None)
    def test_layer_from_uri(self):
        """
        The CrateLayer can also be created by providing an URI that points to
        a CrateDB tarball.
        """
        with urllib.request.urlopen("https://crate.io/versions.json") as response:
            versions = json.loads(response.read().decode())
            version = versions["crate_testing"]

        self.assertGreaterEqual(Version(version), Version("4.5.0"))

        uri = "https://cdn.crate.io/downloads/releases/crate-{}.tar.gz".format(version)
        layer = CrateLayer.from_uri(uri, name="crate-by-uri", http_port=42203)
        self.assertIsInstance(layer, CrateLayer)

    @mock.patch.dict('os.environ', {}, clear=True)
    def test_java_home_env_not_set(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            layer = CrateLayer('java-home-test', tmpdir)
            # JAVA_HOME must not be set to `None`, since it would be interpreted as a
            # string 'None', and therefore intepreted as a path
            self.assertEqual(layer.env['JAVA_HOME'], '')

    @mock.patch.dict('os.environ', {}, clear=True)
    def test_java_home_env_set(self):
        java_home = '/usr/lib/jvm/java-11-openjdk-amd64'
        with tempfile.TemporaryDirectory() as tmpdir:
            os.environ['JAVA_HOME'] = java_home
            layer = CrateLayer('java-home-test', tmpdir)
            self.assertEqual(layer.env['JAVA_HOME'], java_home)

    @mock.patch.dict('os.environ', {}, clear=True)
    def test_java_home_env_override(self):
        java_11_home = '/usr/lib/jvm/java-11-openjdk-amd64'
        java_12_home = '/usr/lib/jvm/java-12-openjdk-amd64'
        with tempfile.TemporaryDirectory() as tmpdir:
            os.environ['JAVA_HOME'] = java_11_home
            layer = CrateLayer('java-home-test', tmpdir, env={'JAVA_HOME': java_12_home})
            self.assertEqual(layer.env['JAVA_HOME'], java_12_home)


class LayerTest(TestCase):

    def test_basic(self):
        """
        This layer starts and stops a ``Crate`` instance on a given host, port,
        a given crate node name and, optionally, a given cluster name::
        """

        port = 44219
        transport_port = 44319

        layer = CrateLayer('crate',
                           crate_home=crate_path(),
                           host='127.0.0.1',
                           port=port,
                           transport_port=transport_port,
                           cluster_name='my_cluster'
                           )

        # The working directory is defined on layer instantiation.
        # It is sometimes required to know it before starting the layer.
        self.assertRegex(layer.wdPath(), r".+/crate_layer/crate")

        # Start the layer.
        layer.start()

        # The urls of the crate servers to be instantiated can be obtained
        # via `crate_servers`.
        self.assertEqual(layer.crate_servers, ["http://127.0.0.1:44219"])

        # Access the CrateDB instance on the HTTP interface.

        http = urllib3.PoolManager()

        stats_uri = "http://127.0.0.1:{0}/".format(port)
        response = http.request('GET', stats_uri)
        self.assertEqual(response.status, 200)

        # The layer can be shutdown using its `stop()` method.
        layer.stop()

    def test_dynamic_http_port(self):
        """
        It is also possible to define a port range instead of a static HTTP port for the layer.

        Crate will start with the first available port in the given range and the test
        layer obtains the chosen port from the startup logs of the Crate process.
        Note, that this feature requires a logging configuration with at least loglevel
        ``INFO`` on ``http``.
        """
        port = '44200-44299'
        layer = CrateLayer('crate', crate_home=crate_path(), port=port)
        layer.start()
        self.assertRegex(layer.crate_servers[0], r"http://127.0.0.1:442\d\d")
        layer.stop()

    def test_default_settings(self):
        """
        Starting a CrateDB layer leaving out optional parameters will apply the following
        defaults.

        The default http port is the first free port in the range of ``4200-4299``,
        the default transport port is the first free port in the range of ``4300-4399``,
        the host defaults to ``127.0.0.1``.

        The command to call is ``bin/crate`` inside the ``crate_home`` path.
        The default config file is ``config/crate.yml`` inside ``crate_home``.
        The default cluster name will be auto generated using the HTTP port.
        """
        layer = CrateLayer('crate_defaults', crate_home=crate_path())
        layer.start()
        self.assertEqual(layer.crate_servers[0], "http://127.0.0.1:4200")
        layer.stop()

    def test_additional_settings(self):
        """
        The ``Crate`` layer can be started with additional settings as well.
        Add a dictionary for keyword argument ``settings`` which contains your settings.
        Those additional setting will override settings given as keyword argument.

        The settings will be handed over to the ``Crate`` process with the ``-C`` flag.
        So the setting ``threadpool.bulk.queue_size: 100`` becomes
        the command line flag: ``-Cthreadpool.bulk.queue_size=100``::
        """
        layer = CrateLayer(
            'custom',
            crate_path(),
            port=44401,
            settings={
                "cluster.graceful_stop.min_availability": "none",
                "http.port": 44402
            }
        )
        layer.start()
        self.assertEqual(layer.crate_servers[0], "http://127.0.0.1:44402")
        self.assertIn("-Ccluster.graceful_stop.min_availability=none", layer.start_cmd)
        layer.stop()

    def test_verbosity(self):
        """
        The test layer hides the standard output of Crate per default. To increase the
        verbosity level the additional keyword argument ``verbose`` needs to be set
        to ``True``::
        """
        layer = CrateLayer('crate',
                           crate_home=crate_path(),
                           verbose=True)
        layer.start()
        self.assertTrue(layer.verbose)
        layer.stop()

    def test_environment_variables(self):
        """
        It is possible to provide environment variables for the ``Crate`` testing
        layer.
        """
        layer = CrateLayer('crate',
                           crate_home=crate_path(),
                           env={"CRATE_HEAP_SIZE": "300m"})

        layer.start()

        sql_uri = layer.crate_servers[0] + "/_sql"

        http = urllib3.PoolManager()
        response = http.urlopen('POST', sql_uri,
                                body='{"stmt": "select heap[\'max\'] from sys.nodes"}')
        json_response = json.loads(response.data.decode('utf-8'))

        self.assertEqual(json_response["rows"][0][0], 314572800)

        layer.stop()

    def test_cluster(self):
        """
        To start a cluster of ``Crate`` instances, give each instance the same
        ``cluster_name``. If you want to start instances on the same machine then
        use value ``_local_`` for ``host`` and give every node different ports::
        """
        cluster_layer1 = CrateLayer(
            'crate1',
            crate_path(),
            host='_local_',
            cluster_name='my_cluster',
        )
        cluster_layer2 = CrateLayer(
            'crate2',
            crate_path(),
            host='_local_',
            cluster_name='my_cluster',
            settings={"discovery.initial_state_timeout": "10s"}
        )

        # If we start both layers, they will, after a small amount of time, find each other
        # and form a cluster.
        cluster_layer1.start()
        cluster_layer2.start()

        # We can verify that by checking the number of nodes a node knows about.
        http = urllib3.PoolManager()

        def num_cluster_nodes(crate_layer):
            sql_uri = crate_layer.crate_servers[0] + "/_sql"
            response = http.urlopen('POST', sql_uri, body='{"stmt":"select count(*) from sys.nodes"}')
            json_response = json.loads(response.data.decode('utf-8'))
            return json_response["rows"][0][0]

        # We might have to wait a moment before the cluster is finally created.
        num_nodes = num_cluster_nodes(cluster_layer1)
        import time
        retries = 0
        while num_nodes < 2:  # pragma: no cover
            time.sleep(1)
            num_nodes = num_cluster_nodes(cluster_layer1)
            retries += 1
            if retries == 30:
                break
        self.assertEqual(num_nodes, 2)

        cluster_layer1.stop()
        cluster_layer2.stop()



# -*- coding: utf-8 -*-

from ..compat.api13 import monkeypatch_amend_select_sa14, monkeypatch_add_connectionfairy_driver_connection
from ..sa_version import SA_1_4, SA_VERSION
from ...test_util import ParametrizedTestCase

# `sql.select()` of SQLAlchemy 1.3 uses old calling semantics,
# but the test cases already need the modern ones.
if SA_VERSION < SA_1_4:
    monkeypatch_amend_select_sa14()
    monkeypatch_add_connectionfairy_driver_connection()

from unittest import TestLoader, TestSuite
from .connection_test import SqlAlchemyConnectionTest
from .dict_test import SqlAlchemyDictTypeTest
from .datetime_test import SqlAlchemyDateAndDateTimeTest
from .compiler_test import SqlAlchemyCompilerTest, SqlAlchemyDDLCompilerTest
from .update_test import SqlAlchemyUpdateTest
from .match_test import SqlAlchemyMatchTest
from .bulk_test import SqlAlchemyBulkTest
from .insert_from_select_test import SqlAlchemyInsertFromSelectTest
from .create_table_test import SqlAlchemyCreateTableTest
from .array_test import SqlAlchemyArrayTypeTest
from .dialect_test import SqlAlchemyDialectTest
from .function_test import SqlAlchemyFunctionTest
from .warnings_test import SqlAlchemyWarningsTest
from .query_caching import SqlAlchemyQueryCompilationCaching


makeSuite = TestLoader().loadTestsFromTestCase


def test_suite_unit():
    tests = TestSuite()
    tests.addTest(makeSuite(SqlAlchemyConnectionTest))
    tests.addTest(makeSuite(SqlAlchemyDictTypeTest))
    tests.addTest(makeSuite(SqlAlchemyDateAndDateTimeTest))
    tests.addTest(makeSuite(SqlAlchemyCompilerTest))
    tests.addTest(makeSuite(SqlAlchemyDDLCompilerTest))
    tests.addTest(ParametrizedTestCase.parametrize(SqlAlchemyCompilerTest, param={"server_version_info": None}))
    tests.addTest(ParametrizedTestCase.parametrize(SqlAlchemyCompilerTest, param={"server_version_info": (4, 0, 12)}))
    tests.addTest(ParametrizedTestCase.parametrize(SqlAlchemyCompilerTest, param={"server_version_info": (4, 1, 10)}))
    tests.addTest(makeSuite(SqlAlchemyUpdateTest))
    tests.addTest(makeSuite(SqlAlchemyMatchTest))
    tests.addTest(makeSuite(SqlAlchemyCreateTableTest))
    tests.addTest(makeSuite(SqlAlchemyBulkTest))
    tests.addTest(makeSuite(SqlAlchemyInsertFromSelectTest))
    tests.addTest(makeSuite(SqlAlchemyInsertFromSelectTest))
    tests.addTest(makeSuite(SqlAlchemyDialectTest))
    tests.addTest(makeSuite(SqlAlchemyFunctionTest))
    tests.addTest(makeSuite(SqlAlchemyArrayTypeTest))
    tests.addTest(makeSuite(SqlAlchemyWarningsTest))
    return tests


def test_suite_integration():
    tests = TestSuite()
    tests.addTest(makeSuite(SqlAlchemyQueryCompilationCaching))
    return tests



# -*- coding: utf-8; -*-
import sys
import warnings
from unittest import TestCase, skipIf

from crate.client.sqlalchemy import SA_1_4, SA_VERSION
from crate.testing.util import ExtraAssertions


class SqlAlchemyWarningsTest(TestCase, ExtraAssertions):
    """
    Verify a few `DeprecationWarning` spots.

    https://docs.python.org/3/library/warnings.html#testing-warnings
    """

    @skipIf(SA_VERSION >= SA_1_4, "There is no deprecation warning for "
                                  "SQLAlchemy 1.3 on higher versions")
    def test_sa13_deprecation_warning(self):
        """
        Verify that a `DeprecationWarning` is issued when running SQLAlchemy 1.3.
        """
        with warnings.catch_warnings(record=True) as w:

            # Cause all warnings to always be triggered.
            warnings.simplefilter("always")

            # Trigger a warning by importing the SQLAlchemy dialect module.
            # Because it already has been loaded, unload it beforehand.
            del sys.modules["crate.client.sqlalchemy"]
            import crate.client.sqlalchemy  # noqa: F401

            # Verify details of the SA13 EOL/deprecation warning.
            self.assertEqual(len(w), 1)
            self.assertIsSubclass(w[-1].category, DeprecationWarning)
            self.assertIn("SQLAlchemy 1.3 is effectively EOL.", str(w[-1].message))

    def test_craty_object_deprecation_warning(self):
        """
        Verify that a `DeprecationWarning` is issued when accessing the deprecated
        module variables `Craty`, and `Object`. The new type is called `ObjectType`.
        """

        with warnings.catch_warnings(record=True) as w:

            # Import the deprecated symbol.
            from crate.client.sqlalchemy.types import Craty  # noqa: F401

            # Verify details of the deprecation warning.
            self.assertEqual(len(w), 1)
            self.assertIsSubclass(w[-1].category, DeprecationWarning)
            self.assertIn("Craty is deprecated and will be removed in future releases. "
                          "Please use ObjectType instead.", str(w[-1].message))

        with warnings.catch_warnings(record=True) as w:

            # Import the deprecated symbol.
            from crate.client.sqlalchemy.types import Object  # noqa: F401

            # Verify details of the deprecation warning.
            self.assertEqual(len(w), 1)
            self.assertIsSubclass(w[-1].category, DeprecationWarning)
            self.assertIn("Object is deprecated and will be removed in future releases. "
                          "Please use ObjectType instead.", str(w[-1].message))



# -*- coding: utf-8; -*-
#
# Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
# license agreements.  See the NOTICE file distributed with this work for
# additional information regarding copyright ownership.  Crate licenses
# this file to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may
# obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
# However, if you have executed another commercial license agreement
# with Crate these terms will supersede the license and you may use the
# software solely pursuant to the terms of the relevant commercial agreement.

import sqlalchemy as sa
from sqlalchemy.dialects.postgresql.base import PGCompiler
from sqlalchemy.sql.crud import (REQUIRED, _create_bind_param,
                                 _extend_values_for_multiparams,
                                 _get_multitable_params,
                                 _get_stmt_parameters_params,
                                 _key_getters_for_crud_column, _scan_cols,
                                 _scan_insert_from_select_cols)

from crate.client.sqlalchemy.compiler import CrateCompiler


class CrateCompilerSA10(CrateCompiler):

    def returning_clause(self, stmt, returning_cols):
        """
        Generate RETURNING clause, PostgreSQL-compatible.
        """
        return PGCompiler.returning_clause(self, stmt, returning_cols)

    def visit_update(self, update_stmt, **kw):
        """
        used to compile <sql.expression.Update> expressions
        Parts are taken from the SQLCompiler base class.
        """

        # [10] CrateDB patch.
        if not update_stmt.parameters and \
                not hasattr(update_stmt, '_crate_specific'):
            return super().visit_update(update_stmt, **kw)

        self.isupdate = True

        extra_froms = update_stmt._extra_froms

        text = 'UPDATE '

        if update_stmt._prefixes:
            text += self._generate_prefixes(update_stmt,
                                            update_stmt._prefixes, **kw)

        table_text = self.update_tables_clause(update_stmt, update_stmt.table,
                                               extra_froms, **kw)

        dialect_hints = None
        if update_stmt._hints:
            dialect_hints, table_text = self._setup_crud_hints(
                update_stmt, table_text
            )

        # [10] CrateDB patch.
        crud_params = _get_crud_params(self, update_stmt, **kw)

        text += table_text

        text += ' SET '

        # [10] CrateDB patch begin.
        include_table = \
            extra_froms and self.render_table_with_column_in_update_from

        set_clauses = []

        for k, v in crud_params:
            clause = k._compiler_dispatch(self,
                                          include_table=include_table) + \
                ' = ' + v
            set_clauses.append(clause)

        for k, v in update_stmt.parameters.items():
            if isinstance(k, str) and '[' in k:
                bindparam = sa.sql.bindparam(k, v)
                set_clauses.append(k + ' = ' + self.process(bindparam))

        text += ', '.join(set_clauses)
        # [10] CrateDB patch end.

        if self.returning or update_stmt._returning:
            if not self.returning:
                self.returning = update_stmt._returning
            if self.returning_precedes_values:
                text += " " + self.returning_clause(
                    update_stmt, self.returning)

        if extra_froms:
            extra_from_text = self.update_from_clause(
                update_stmt,
                update_stmt.table,
                extra_froms,
                dialect_hints,
                **kw)
            if extra_from_text:
                text += " " + extra_from_text

        if update_stmt._whereclause is not None:
            t = self.process(update_stmt._whereclause)
            if t:
                text += " WHERE " + t

        limit_clause = self.update_limit_clause(update_stmt)
        if limit_clause:
            text += " " + limit_clause

        if self.returning and not self.returning_precedes_values:
            text += " " + self.returning_clause(
                update_stmt, self.returning)

        return text


def _get_crud_params(compiler, stmt, **kw):
    """create a set of tuples representing column/string pairs for use
    in an INSERT or UPDATE statement.

    Also generates the Compiled object's postfetch, prefetch, and
    returning column collections, used for default handling and ultimately
    populating the ResultProxy's prefetch_cols() and postfetch_cols()
    collections.

    """

    compiler.postfetch = []
    compiler.insert_prefetch = []
    compiler.update_prefetch = []
    compiler.returning = []

    # no parameters in the statement, no parameters in the
    # compiled params - return binds for all columns
    if compiler.column_keys is None and stmt.parameters is None:
        return [
            (c, _create_bind_param(compiler, c, None, required=True))
            for c in stmt.table.columns
        ]

    if stmt._has_multi_parameters:
        stmt_parameters = stmt.parameters[0]
    else:
        stmt_parameters = stmt.parameters

    # getters - these are normally just column.key,
    # but in the case of mysql multi-table update, the rules for
    # .key must conditionally take tablename into account
    (
        _column_as_key,
        _getattr_col_key,
        _col_bind_name,
    ) = _key_getters_for_crud_column(compiler, stmt)

    # if we have statement parameters - set defaults in the
    # compiled params
    if compiler.column_keys is None:
        parameters = {}
    else:
        parameters = dict(
            (_column_as_key(key), REQUIRED)
            for key in compiler.column_keys
            if not stmt_parameters or key not in stmt_parameters
        )

    # create a list of column assignment clauses as tuples
    values = []

    if stmt_parameters is not None:
        _get_stmt_parameters_params(
            compiler, parameters, stmt_parameters, _column_as_key, values, kw
        )

    check_columns = {}

    # special logic that only occurs for multi-table UPDATE
    # statements
    if compiler.isupdate and stmt._extra_froms and stmt_parameters:
        _get_multitable_params(
            compiler,
            stmt,
            stmt_parameters,
            check_columns,
            _col_bind_name,
            _getattr_col_key,
            values,
            kw,
        )

    if compiler.isinsert and stmt.select_names:
        _scan_insert_from_select_cols(
            compiler,
            stmt,
            parameters,
            _getattr_col_key,
            _column_as_key,
            _col_bind_name,
            check_columns,
            values,
            kw,
        )
    else:
        _scan_cols(
            compiler,
            stmt,
            parameters,
            _getattr_col_key,
            _column_as_key,
            _col_bind_name,
            check_columns,
            values,
            kw,
        )

    # [10] CrateDB patch.
    #
    # This sanity check performed by SQLAlchemy currently needs to be
    # deactivated in order to satisfy the rewriting logic of the CrateDB
    # dialect in `rewrite_update` and `visit_update`.
    #
    # It can be quickly reproduced by activating this section and running the
    # test cases::
    #
    #   ./bin/test -vvvv -t dict_test
    #
    # That croaks like::
    #
    #   sqlalchemy.exc.CompileError: Unconsumed column names: characters_name, data['nested']
    #
    # TODO: Investigate why this is actually happening and eventually mitigate
    #       the root cause.
    """
    if parameters and stmt_parameters:
        check = (
            set(parameters)
            .intersection(_column_as_key(k) for k in stmt_parameters)
            .difference(check_columns)
        )
        if check:
            raise exc.CompileError(
                "Unconsumed column names: %s"
                % (", ".join("%s" % c for c in check))
            )
    """

    if stmt._has_multi_parameters:
        values = _extend_values_for_multiparams(compiler, stmt, values, kw)

    return values



# -*- coding: utf-8; -*-
#
# Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
# license agreements.  See the NOTICE file distributed with this work for
# additional information regarding copyright ownership.  Crate licenses
# this file to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may
# obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
# However, if you have executed another commercial license agreement
# with Crate these terms will supersede the license and you may use the
# software solely pursuant to the terms of the relevant commercial agreement.

from unittest import TestCase
import sqlalchemy as sa
from sqlalchemy.exc import NoSuchModuleError


class SqlAlchemyConnectionTest(TestCase):

    def test_connection_server_uri_unknown_sa_plugin(self):
        with self.assertRaises(NoSuchModuleError):
            sa.create_engine("foobar://otherhost:19201")

    def test_default_connection(self):
        engine = sa.create_engine('crate://')
        conn = engine.raw_connection()
        self.assertEqual("<Connection <Client ['http://127.0.0.1:4200']>>",
                         repr(conn.driver_connection))
        conn.close()
        engine.dispose()

    def test_connection_server_uri_http(self):
        engine = sa.create_engine(
            "crate://otherhost:19201")
        conn = engine.raw_connection()
        self.assertEqual("<Connection <Client ['http://otherhost:19201']>>",
                         repr(conn.driver_connection))
        conn.close()
        engine.dispose()

    def test_connection_server_uri_https(self):
        engine = sa.create_engine(
            "crate://otherhost:19201/?ssl=true")
        conn = engine.raw_connection()
        self.assertEqual("<Connection <Client ['https://otherhost:19201']>>",
                         repr(conn.driver_connection))
        conn.close()
        engine.dispose()

    def test_connection_server_uri_invalid_port(self):
        with self.assertRaises(ValueError) as context:
            sa.create_engine("crate://foo:bar")
        self.assertIn("invalid literal for int() with base 10: 'bar'", str(context.exception))

    def test_connection_server_uri_https_with_trusted_user(self):
        engine = sa.create_engine(
            "crate://foo@otherhost:19201/?ssl=true")
        conn = engine.raw_connection()
        self.assertEqual("<Connection <Client ['https://otherhost:19201']>>",
                         repr(conn.driver_connection))
        self.assertEqual(conn.driver_connection.client.username, "foo")
        self.assertEqual(conn.driver_connection.client.password, None)
        conn.close()
        engine.dispose()

    def test_connection_server_uri_https_with_credentials(self):
        engine = sa.create_engine(
            "crate://foo:bar@otherhost:19201/?ssl=true")
        conn = engine.raw_connection()
        self.assertEqual("<Connection <Client ['https://otherhost:19201']>>",
                         repr(conn.driver_connection))
        self.assertEqual(conn.driver_connection.client.username, "foo")
        self.assertEqual(conn.driver_connection.client.password, "bar")
        conn.close()
        engine.dispose()

    def test_connection_server_uri_parameter_timeout(self):
        engine = sa.create_engine(
            "crate://otherhost:19201/?timeout=42.42")
        conn = engine.raw_connection()
        self.assertEqual(conn.driver_connection.client._pool_kw["timeout"], 42.42)
        conn.close()
        engine.dispose()

    def test_connection_server_uri_parameter_pool_size(self):
        engine = sa.create_engine(
            "crate://otherhost:19201/?pool_size=20")
        conn = engine.raw_connection()
        self.assertEqual(conn.driver_connection.client._pool_kw["maxsize"], 20)
        conn.close()
        engine.dispose()

    def test_connection_multiple_server_http(self):
        engine = sa.create_engine(
            "crate://", connect_args={
                'servers': ['localhost:4201', 'localhost:4202']
            }
        )
        conn = engine.raw_connection()
        self.assertEqual(
            "<Connection <Client ['http://localhost:4201', " +
            "'http://localhost:4202']>>",
            repr(conn.driver_connection))
        conn.close()
        engine.dispose()

    def test_connection_multiple_server_https(self):
        engine = sa.create_engine(
            "crate://", connect_args={
                'servers': ['localhost:4201', 'localhost:4202'],
                'ssl': True,
            }
        )
        conn = engine.raw_connection()
        self.assertEqual(
            "<Connection <Client ['https://localhost:4201', " +
            "'https://localhost:4202']>>",
            repr(conn.driver_connection))
        conn.close()
        engine.dispose()



# -*- coding: utf-8; -*-
#
# Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
# license agreements.  See the NOTICE file distributed with this work for
# additional information regarding copyright ownership.  Crate licenses
# this file to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may
# obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
# However, if you have executed another commercial license agreement
# with Crate these terms will supersede the license and you may use the
# software solely pursuant to the terms of the relevant commercial agreement.
import warnings
from textwrap import dedent
from unittest import mock, skipIf, TestCase
from unittest.mock import MagicMock, patch

from crate.client.cursor import Cursor
from crate.client.sqlalchemy.compiler import crate_before_execute

import sqlalchemy as sa
from sqlalchemy.sql import text, Update

from crate.testing.util import ExtraAssertions

try:
    from sqlalchemy.orm import declarative_base
except ImportError:
    from sqlalchemy.ext.declarative import declarative_base

from crate.client.sqlalchemy.sa_version import SA_VERSION, SA_1_4, SA_2_0
from crate.client.sqlalchemy.types import ObjectType
from crate.client.test_util import ParametrizedTestCase

from crate.testing.settings import crate_host


class SqlAlchemyCompilerTest(ParametrizedTestCase, ExtraAssertions):

    def setUp(self):
        self.crate_engine = sa.create_engine('crate://')
        if isinstance(self.param, dict) and "server_version_info" in self.param:
            server_version_info = self.param["server_version_info"]
            self.crate_engine.dialect.server_version_info = server_version_info
        self.sqlite_engine = sa.create_engine('sqlite://')
        self.metadata = sa.MetaData()
        self.mytable = sa.Table('mytable', self.metadata,
                                sa.Column('name', sa.String),
                                sa.Column('data', ObjectType))

        self.update = Update(self.mytable).where(text('name=:name'))
        self.values = [{'name': 'crate'}]
        self.values = (self.values, )

    def test_sqlite_update_not_rewritten(self):
        clauseelement, multiparams, params = crate_before_execute(
            self.sqlite_engine, self.update, self.values, {}
        )

        self.assertFalse(hasattr(clauseelement, '_crate_specific'))

    def test_crate_update_rewritten(self):
        clauseelement, multiparams, params = crate_before_execute(
            self.crate_engine, self.update, self.values, {}
        )

        self.assertTrue(hasattr(clauseelement, '_crate_specific'))

    def test_bulk_update_on_builtin_type(self):
        """
        The "before_execute" hook in the compiler doesn't get
        access to the parameters in case of a bulk update. It
        should not try to optimize any parameters.
        """
        data = ({},)
        clauseelement, multiparams, params = crate_before_execute(
            self.crate_engine, self.update, data, None
        )

        self.assertFalse(hasattr(clauseelement, '_crate_specific'))

    def test_select_with_ilike_no_escape(self):
        """
        Verify the compiler uses CrateDB's native `ILIKE` method.
        """
        selectable = self.mytable.select().where(self.mytable.c.name.ilike("%foo%"))
        statement = str(selectable.compile(bind=self.crate_engine))
        if self.crate_engine.dialect.has_ilike_operator():
            self.assertEqual(statement, dedent("""
                SELECT mytable.name, mytable.data 
                FROM mytable 
                WHERE mytable.name ILIKE ?
            """).strip())  # noqa: W291
        else:
            self.assertEqual(statement, dedent("""
                SELECT mytable.name, mytable.data 
                FROM mytable 
                WHERE lower(mytable.name) LIKE lower(?)
            """).strip())  # noqa: W291

    def test_select_with_not_ilike_no_escape(self):
        """
        Verify the compiler uses CrateDB's native `ILIKE` method.
        """
        selectable = self.mytable.select().where(self.mytable.c.name.notilike("%foo%"))
        statement = str(selectable.compile(bind=self.crate_engine))
        if SA_VERSION < SA_1_4 or not self.crate_engine.dialect.has_ilike_operator():
            self.assertEqual(statement, dedent("""
                SELECT mytable.name, mytable.data 
                FROM mytable 
                WHERE lower(mytable.name) NOT LIKE lower(?)
            """).strip())  # noqa: W291
        else:
            self.assertEqual(statement, dedent("""
                SELECT mytable.name, mytable.data 
                FROM mytable 
                WHERE mytable.name NOT ILIKE ?
            """).strip())  # noqa: W291

    def test_select_with_ilike_and_escape(self):
        """
        Verify the compiler fails when using CrateDB's native `ILIKE` method together with `ESCAPE`.
        """

        selectable = self.mytable.select().where(self.mytable.c.name.ilike("%foo%", escape='\\'))
        with self.assertRaises(NotImplementedError) as cmex:
            selectable.compile(bind=self.crate_engine)
        self.assertEqual(str(cmex.exception), "Unsupported feature: ESCAPE is not supported")

    @skipIf(SA_VERSION < SA_1_4, "SQLAlchemy 1.3 and earlier do not support native `NOT ILIKE` compilation")
    def test_select_with_not_ilike_and_escape(self):
        """
        Verify the compiler fails when using CrateDB's native `ILIKE` method together with `ESCAPE`.
        """

        selectable = self.mytable.select().where(self.mytable.c.name.notilike("%foo%", escape='\\'))
        with self.assertRaises(NotImplementedError) as cmex:
            selectable.compile(bind=self.crate_engine)
        self.assertEqual(str(cmex.exception), "Unsupported feature: ESCAPE is not supported")

    def test_select_with_offset(self):
        """
        Verify the `CrateCompiler.limit_clause` method, with offset.
        """
        selectable = self.mytable.select().offset(5)
        statement = str(selectable.compile(bind=self.crate_engine))
        if SA_VERSION >= SA_1_4:
            self.assertEqual(statement, "SELECT mytable.name, mytable.data \nFROM mytable\n LIMIT ALL OFFSET ?")
        else:
            self.assertEqual(statement, "SELECT mytable.name, mytable.data \nFROM mytable \n LIMIT ALL OFFSET ?")

    def test_select_with_limit(self):
        """
        Verify the `CrateCompiler.limit_clause` method, with limit.
        """
        selectable = self.mytable.select().limit(42)
        statement = str(selectable.compile(bind=self.crate_engine))
        self.assertEqual(statement, "SELECT mytable.name, mytable.data \nFROM mytable \n LIMIT ?")

    def test_select_with_offset_and_limit(self):
        """
        Verify the `CrateCompiler.limit_clause` method, with offset and limit.
        """
        selectable = self.mytable.select().offset(5).limit(42)
        statement = str(selectable.compile(bind=self.crate_engine))
        self.assertEqual(statement, "SELECT mytable.name, mytable.data \nFROM mytable \n LIMIT ? OFFSET ?")

    def test_insert_multivalues(self):
        """
        Verify that "in-place multirow inserts" aka. "multivalues inserts" aka.
        the `supports_multivalues_insert` dialect feature works.

        When this feature is not enabled, using it will raise an error:

            CompileError: The 'crate' dialect with current database version
            settings does not support in-place multirow inserts

        > The Insert construct also supports being passed a list of dictionaries
        > or full-table-tuples, which on the server will render the less common
        > SQL syntax of "multiple values" - this syntax is supported on backends
        > such as SQLite, PostgreSQL, MySQL, but not necessarily others.

        > It is essential to note that passing multiple values is NOT the same
        > as using traditional `executemany()` form. The above syntax is a special
        > syntax not typically used. To emit an INSERT statement against
        > multiple rows, the normal method is to pass a multiple values list to
        > the `Connection.execute()` method, which is supported by all database
        > backends and is generally more efficient for a very large number of
        > parameters.

        - https://docs.sqlalchemy.org/core/dml.html#sqlalchemy.sql.expression.Insert.values.params.*args
        """
        records = [{"name": f"foo_{i}"} for i in range(3)]
        insertable = self.mytable.insert().values(records)
        statement = str(insertable.compile(bind=self.crate_engine))
        self.assertEqual(statement, "INSERT INTO mytable (name) VALUES (?), (?), (?)")

    @skipIf(SA_VERSION < SA_2_0, "SQLAlchemy 1.x does not support the 'insertmanyvalues' dialect feature")
    def test_insert_manyvalues(self):
        """
        Verify the `use_insertmanyvalues` and `use_insertmanyvalues_wo_returning` dialect features.

        > For DML statements such as "INSERT", "UPDATE" and "DELETE", we can
        > send multiple parameter sets to the `Connection.execute()` method by
        > passing a list of dictionaries instead of a single dictionary, which
        > indicates that the single SQL statement should be invoked multiple
        > times, once for each parameter set. This style of execution is known
        > as "executemany".

        > A key characteristic of "insertmanyvalues" is that the size of the INSERT
        > statement is limited on a fixed max number of "values" clauses as well as
        > a dialect-specific fixed total number of bound parameters that may be
        > represented in one INSERT statement at a time.
        > When the number of parameter dictionaries given exceeds a fixed limit [...],
        > multiple INSERT statements will be invoked within the scope of a single
        > `Connection.execute()` call, each of which accommodate for a portion of the
        > parameter dictionaries, referred towards as a "batch".

        - https://docs.sqlalchemy.org/tutorial/dbapi_transactions.html#tutorial-multiple-parameters
        - https://docs.sqlalchemy.org/glossary.html#term-executemany
        - https://docs.sqlalchemy.org/core/connections.html#engine-insertmanyvalues
        - https://docs.sqlalchemy.org/core/connections.html#controlling-the-batch-size
        """

        # Don't truncate unittest's diff output on `assertListEqual`.
        self.maxDiff = None

        # Five records with a batch size of two should produce three `INSERT` statements.
        record_count = 5
        batch_size = 2

        # Prepare input data and verify insert statement.
        records = [{"name": f"foo_{i}"} for i in range(record_count)]
        insertable = self.mytable.insert()
        statement = str(insertable.compile(bind=self.crate_engine))
        self.assertEqual(statement, "INSERT INTO mytable (name, data) VALUES (?, ?)")

        with mock.patch("crate.client.http.Client.sql", autospec=True, return_value={"cols": []}) as client_mock:

            with self.crate_engine.begin() as conn:
                # Adjust page size on a per-connection level.
                conn.execution_options(insertmanyvalues_page_size=batch_size)
                conn.execute(insertable, parameters=records)

        # Verify that input data has been batched correctly.
        self.assertListEqual(client_mock.mock_calls, [
            mock.call(mock.ANY, 'INSERT INTO mytable (name) VALUES (?), (?)', ('foo_0', 'foo_1'), None),
            mock.call(mock.ANY, 'INSERT INTO mytable (name) VALUES (?), (?)', ('foo_2', 'foo_3'), None),
            mock.call(mock.ANY, 'INSERT INTO mytable (name) VALUES (?)', ('foo_4', ), None),
        ])

    def test_for_update(self):
        """
        Verify the `CrateCompiler.for_update_clause` method to
        omit the clause, since CrateDB does not support it.
        """

        with warnings.catch_warnings(record=True) as w:

            # By default, warnings from a loop will only be emitted once.
            # This scenario tests exactly this behaviour, to verify logs
            # don't get flooded.
            warnings.simplefilter("once")

            selectable = self.mytable.select().with_for_update()
            _ = str(selectable.compile(bind=self.crate_engine))

            selectable = self.mytable.select().with_for_update()
            statement = str(selectable.compile(bind=self.crate_engine))

        # Verify SQL statement.
        self.assertEqual(statement, "SELECT mytable.name, mytable.data \nFROM mytable")

        # Verify if corresponding warning is emitted, once.
        self.assertEqual(len(w), 1)
        self.assertIsSubclass(w[-1].category, UserWarning)
        self.assertIn("CrateDB does not support the 'INSERT ... FOR UPDATE' clause, "
                      "it will be omitted when generating SQL statements.", str(w[-1].message))


FakeCursor = MagicMock(name='FakeCursor', spec=Cursor)


class CompilerTestCase(TestCase):
    """
    A base class for providing mocking infrastructure to validate the DDL compiler.
    """

    def setUp(self):
        self.engine = sa.create_engine(f"crate://{crate_host}")
        self.metadata = sa.MetaData(schema="testdrive")
        self.session = sa.orm.Session(bind=self.engine)
        self.setup_mock()

    def setup_mock(self):
        """
        Set up a fake cursor, in order to intercept query execution.
        """

        self.fake_cursor = MagicMock(name="fake_cursor")
        FakeCursor.return_value = self.fake_cursor

        self.executed_statement = None
        self.fake_cursor.execute = self.execute_wrapper

    def execute_wrapper(self, query, *args, **kwargs):
        """
        Receive the SQL query expression, and store it.
        """
        self.executed_statement = query
        return self.fake_cursor


@patch('crate.client.connection.Cursor', FakeCursor)
class SqlAlchemyDDLCompilerTest(CompilerTestCase, ExtraAssertions):
    """
    Verify a few scenarios regarding the DDL compiler.
    """

    def test_ddl_with_foreign_keys(self):
        """
        Verify the CrateDB dialect properly ignores foreign key constraints.
        """

        Base = declarative_base(metadata=self.metadata)

        class RootStore(Base):
            """The main store."""

            __tablename__ = "root"

            id = sa.Column(sa.Integer, primary_key=True)
            name = sa.Column(sa.String)

            items = sa.orm.relationship(
                "ItemStore",
                back_populates="root",
                passive_deletes=True,
            )

        class ItemStore(Base):
            """The auxiliary store."""

            __tablename__ = "item"

            id = sa.Column(sa.Integer, primary_key=True)
            name = sa.Column(sa.String)
            root_id = sa.Column(
                sa.Integer,
                sa.ForeignKey(
                    f"{RootStore.__tablename__}.id",
                    ondelete="CASCADE",
                ),
            )
            root = sa.orm.relationship(RootStore, back_populates="items")

        with warnings.catch_warnings(record=True) as w:

            # Cause all warnings to always be triggered.
            warnings.simplefilter("always")

            # Verify SQL DDL statement.
            self.metadata.create_all(self.engine, tables=[RootStore.__table__], checkfirst=False)
            self.assertEqual(self.executed_statement, dedent("""
                CREATE TABLE testdrive.root (
                \tid INT NOT NULL, 
                \tname STRING, 
                \tPRIMARY KEY (id)
                )
    
            """))  # noqa: W291, W293

            # Verify SQL DDL statement.
            self.metadata.create_all(self.engine, tables=[ItemStore.__table__], checkfirst=False)
            self.assertEqual(self.executed_statement, dedent("""
                CREATE TABLE testdrive.item (
                \tid INT NOT NULL, 
                \tname STRING, 
                \troot_id INT, 
                \tPRIMARY KEY (id)
                )
    
            """))  # noqa: W291, W293

        # Verify if corresponding warning is emitted.
        self.assertEqual(len(w), 1)
        self.assertIsSubclass(w[-1].category, UserWarning)
        self.assertIn("CrateDB does not support foreign key constraints, "
                      "they will be omitted when generating DDL statements.", str(w[-1].message))

    def test_ddl_with_unique_key(self):
        """
        Verify the CrateDB dialect properly ignores unique key constraints.
        """

        Base = declarative_base(metadata=self.metadata)

        class FooBar(Base):
            """The entity."""

            __tablename__ = "foobar"

            id = sa.Column(sa.Integer, primary_key=True)
            name = sa.Column(sa.String, unique=True)

        with warnings.catch_warnings(record=True) as w:

            # Cause all warnings to always be triggered.
            warnings.simplefilter("always")

            # Verify SQL DDL statement.
            self.metadata.create_all(self.engine, tables=[FooBar.__table__], checkfirst=False)
            self.assertEqual(self.executed_statement, dedent("""
                CREATE TABLE testdrive.foobar (
                \tid INT NOT NULL, 
                \tname STRING, 
                \tPRIMARY KEY (id)
                )
    
            """))  # noqa: W291, W293

        # Verify if corresponding warning is emitted.
        self.assertEqual(len(w), 1)
        self.assertIsSubclass(w[-1].category, UserWarning)
        self.assertIn("CrateDB does not support unique constraints, "
                      "they will be omitted when generating DDL statements.", str(w[-1].message))



# -*- coding: utf-8; -*-
#
# Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
# license agreements.  See the NOTICE file distributed with this work for
# additional information regarding copyright ownership.  Crate licenses
# this file to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may
# obtain a copy of the License at
#
#   https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
# However, if you have executed another commercial license agreement
# with Crate these terms will supersede the license and you may use the
# software solely pursuant to the terms of the relevant commercial agreement.

"""
Compatibility module for running a subset of SQLAlchemy 2.0 programs on
SQLAlchemy 1.3. By using monkey-patching, it can do two things:

1. Add the `exec_driver_sql` method to SA's `Connection` and `Engine`.
2. Amend the `sql.select` function to accept the calling semantics of
   the modern variant.

Reason: `exec_driver_sql` gets used within the CrateDB dialect already,
and the new calling semantics of `sql.select` already get used within
many of the test cases already. Please note that the patch for
`sql.select` is only applied when running the test suite.
"""

import collections.abc as collections_abc

from sqlalchemy import exc
from sqlalchemy.sql import Select
from sqlalchemy.sql import select as original_select
from sqlalchemy.util import immutabledict


# `_distill_params_20` copied from SA14's `sqlalchemy.engine.{base,util}`.
_no_tuple = ()
_no_kw = immutabledict()


def _distill_params_20(params):
    if params is None:
        return _no_tuple, _no_kw
    elif isinstance(params, list):
        # collections_abc.MutableSequence): # avoid abc.__instancecheck__
        if params and not isinstance(params[0], (collections_abc.Mapping, tuple)):
            raise exc.ArgumentError(
                "List argument must consist only of tuples or dictionaries"
            )

        return (params,), _no_kw
    elif isinstance(
        params,
        (tuple, dict, immutabledict),
        # only do abc.__instancecheck__ for Mapping after we've checked
        # for plain dictionaries and would otherwise raise
    ) or isinstance(params, collections_abc.Mapping):
        return (params,), _no_kw
    else:
        raise exc.ArgumentError("mapping or sequence expected for parameters")


def exec_driver_sql(self, statement, parameters=None, execution_options=None):
    """
    Adapter for `exec_driver_sql`, which is available since SA14, for SA13.
    """
    if execution_options is not None:
        raise ValueError(
            "SA13 backward-compatibility: "
            "`exec_driver_sql` does not support `execution_options`"
        )
    args_10style, kwargs_10style = _distill_params_20(parameters)
    return self.execute(statement, *args_10style, **kwargs_10style)


def monkeypatch_add_exec_driver_sql():
    """
    Transparently add SA14's `exec_driver_sql()` method to SA13.

    AttributeError: 'Connection' object has no attribute 'exec_driver_sql'
    AttributeError: 'Engine' object has no attribute 'exec_driver_sql'
    """
    from sqlalchemy.engine.base import Connection, Engine

    # Add `exec_driver_sql` method to SA's `Connection` and `Engine` classes.
    Connection.exec_driver_sql = exec_driver_sql
    Engine.exec_driver_sql = exec_driver_sql


def select_sa14(*columns, **kw) -> Select:
    """
    Adapt SA14/SA20's calling semantics of `sql.select()` to SA13.

    With SA20, `select()` no longer accepts varied constructor arguments, only
    the "generative" style of `select()` will be supported. The list of columns
    / tables to select from should be passed positionally.

    Derived from https://github.com/sqlalchemy/alembic/blob/b1fad6b6/alembic/util/sqla_compat.py#L557-L558

    sqlalchemy.exc.ArgumentError: columns argument to select() must be a Python list or other iterable
    """
    if isinstance(columns, tuple) and isinstance(columns[0], list):
        if "whereclause" in kw:
            raise ValueError(
                "SA13 backward-compatibility: "
                "`whereclause` is both in kwargs and columns tuple"
            )
        columns, whereclause = columns
        kw["whereclause"] = whereclause
    return original_select(columns, **kw)


def monkeypatch_amend_select_sa14():
    """
    Make SA13's `sql.select()` transparently accept calling semantics of SA14
    and SA20, by swapping in the newer variant of `select_sa14()`.

    This supports the test suite of `crate-python`, because it already uses the
    modern calling semantics.
    """
    import sqlalchemy

    sqlalchemy.select = select_sa14
    sqlalchemy.sql.select = select_sa14
    sqlalchemy.sql.expression.select = select_sa14


@property
def connectionfairy_driver_connection_sa14(self):
    """The connection object as returned by the driver after a connect.

    .. versionadded:: 1.4.24

    .. seealso::

        :attr:`._ConnectionFairy.dbapi_connection`

        :attr:`._ConnectionRecord.driver_connection`

        :ref:`faq_dbapi_connection`

    """
    return self.connection


def monkeypatch_add_connectionfairy_driver_connection():
    import sqlalchemy.pool.base
    sqlalchemy.pool.base._ConnectionFairy.driver_connection = connectionfairy_driver_connection_sa14
